function ix = boundaryFaceIndices(G, direction, i1, i2)
%Retrieve face indices belonging to subset of global outer faces.
%
% SYNOPSIS:
%   ix = boundaryFaceIndices(G, side, i1, i2)
%
% PARAMETERS:
%   G     - Grid data structure.
%
%   side  - Global side from which to extract face indices.
%           Must be one of {'LEFT', 'RIGHT', 'FRONT', 'BACK', ...
%                           'BOTTOM', 'TOP'}
%
%   i1,i2 - Index ranges for local axes one and two, respectively.
%
% RETURNS:
%   ix    - Required face indices.
%
% SEE ALSO:
%   fluxside, pside.
%
% NOTE:
%   This function is mainly intended for internal use by functions fluxside
%   and pside.

%{
Copyright 2009 SINTEF ICT, Applied Mathematics.

This file is part of The MATLAB Reservoir Simulation Toolbox (MRST).

MRST is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MRST is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MRST.  If not, see <http://www.gnu.org/licenses/>.
%}

% $Id: boundaryFaceIndices.m 1956 2009-03-31 11:06:38Z bska $

st = dbstack(1);
try
   caller = st(1).name;
catch  %#ok
   caller = 'BASE:BoundaryConditions';
end

if isempty(i1) || isempty(i2),
   error([caller, ':IndexSpec:Empty'], ...
         'Empty local index range for boundary condition.');
end

i1 = reshape(i1, [], 1);
i2 = reshape(i2, [], 1);

%% Extract all faces of cells within the given subset
[cells, ft, isOutF] = bdryCellsSubset(G, direction, i1, i2, caller);

fIX   = cumsum([0; double(G.cells.numFaces)]);
hfIX  = mcolon(fIX(cells) + 1, fIX(cells + 1));
faces = G.cellFaces(hfIX, 1);
tags  = G.cellFaces(hfIX, 2);

%% Extract only those faces which have the required tag
ix = faces(isOutF(faces) & tags == ft);


%--------------------------------------------------------------------------
% Helpers follow.
%--------------------------------------------------------------------------


function [cells, faceTag, isOutF] = bdryCellsSubset(G, direction, i1, i2, caller)
%% Determine which indices and face tags to look for
switch direction,
   case 'LEFT',
      % I == min(I)
      [d1, d2, faceTag] = deal(2, 3, 1);
   case 'RIGHT',
      % I == max(I)
      [d1, d2, faceTag] = deal(2, 3, 2);
   case 'BACK',
      % J == min(J)
      [d1, d2, faceTag] = deal(1, 3, 3);
   case 'FRONT',
      % J == max(J)
      [d1, d2, faceTag] = deal(1, 3, 4);
   case 'TOP',
      % K == min(K)
      [d1, d2, faceTag] = deal(1, 2, 5);
   case 'BOTTOM',
      % K == max(K)
      [d1, d2, faceTag] = deal(1, 2, 6);
   otherwise
      error([caller, ':Side:Unknown'],                  ...
            'Boundary side ''%s'' not supported in %s', ...
            direction, caller);
end

%% Determine uniqe outer cells (i.e., cells on boundary)
isOutF = any(double(G.faces.neighbors) == 0, 2);
cells  = find(accumarray(sum(G.faces.neighbors(isOutF,:), 2), 1) > 0);

%% Determine logical indices of these cells
% Assume we will only be called for logically Cartesian grids for which the
% fields 'G.cartDims' and 'G.cells.indexMap' are present.
%
[cIJK{1:3}] = ind2sub(reshape(G.cartDims, 1, []), ...
                      double(G.cells.indexMap(cells)));
cIJK        = [cIJK{:}];

%% Determine whether or not a given cell is within the required subset
inSubSet = ismember(cIJK(:,d1), i1) & ismember(cIJK(:,d2), i2);

%% Extract required cell subset
cells = cells(inSubSet);

