function [X, Y, Z] = buildCornerPtNodes(grdecl)
%BUILDCORNERPTNODES Construct physical nodal coordinates for CP grid.
%
% SYNOPSIS:
%   [X, Y, Z] = buildCornerPtNodes(grdecl)
%
% PARAMETERS:
%   grdecl  - Eclipse file output structure as defined by readGRDECL.
%             Must contain at least the fields 'cartDims', 'COORD' and
%             'ZCORN'.
%
% RETURNS:
%   X, Y, Z - Size 2*grdecl.cartDims arrays of 'x', 'y' and 'z' physical
%             nodal coordinates, respectively.
%
% EXAMPLE:
%   gridfile  = [DATADIR, filesep, 'case.grdecl'];
%   grdecl    = readGRDECL(gridfile);
%   [X, Y, Z] = buildCornerPtNodes(grdecl);
%
% SEE ALSO:
%   readGRDECL.

%{
Copyright 2009 SINTEF ICT, Applied Mathematics.

This file is part of The MATLAB Reservoir Simulation Toolbox (MRST).

MRST is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MRST is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MRST.  If not, see <http://www.gnu.org/licenses/>.
%}

% $Id: buildCornerPtNodes.m 2138 2009-05-04 17:09:29Z bska $

try
   %% Recover logical (cell) dimension of grid
   nx = grdecl.cartDims(1);
   ny = grdecl.cartDims(2);
   nz = grdecl.cartDims(3);

   %% Enumerate individual pillars in grid
   % Pillars 'p1', 'p2', 'p3' and 'p4' bound individual cells in grid.
   pillarIx = reshape(1 : (nx+1)*(ny+1), [nx+1, ny+1]);

   p1 = pillarIx(1:nx,   1:ny);
   p2 = pillarIx(2:nx+1, 1:ny);
   p3 = pillarIx(1:nx,   2:ny+1);
   p4 = pillarIx(2:nx+1, 2:ny+1);
   clear pillarIx

   shape = 2 .* [nx, ny, nz];
   lineIx = zeros(shape);
   lineIx(1:2:2*nx, 1:2:2*ny, :) = repmat(p1, [1, 1, 2*nz]);
   lineIx(2:2:2*nx, 1:2:2*ny, :) = repmat(p2, [1, 1, 2*nz]);
   lineIx(1:2:2*nx, 2:2:2*ny, :) = repmat(p3, [1, 1, 2*nz]);
   lineIx(2:2:2*nx, 2:2:2*ny, :) = repmat(p4, [1, 1, 2*nz]);
   clear p1 p2 p3 p4

   lines = reshape(grdecl.COORD, 6, []) .';
   lines = lines(lineIx(:), :); 
   clear lineIx

   %% Recover physical nodal coordinates along pillars
   propFac = (grdecl.ZCORN(:) - lines(:,3)) ./ (lines(:,6) - lines(:,3));
   xCoords = lines(:,1) + propFac.*(lines(:,4) - lines(:,1));
   yCoords = lines(:,2) + propFac.*(lines(:,5) - lines(:,2));
   clear propFac lines

   %% Reshape to form more amenable to further processing
   X = reshape(xCoords,      shape); clear xCoords;
   Y = reshape(yCoords,      shape); clear yCoords;
   Z = reshape(grdecl.ZCORN, shape);

%% Error handling in case 'grdecl' is invalid.
catch excpt
   if strcmp(excpt.identifier, 'MATLAB:nonExistentField'),
      error('buildCornerPtNodes:Input:Invalid', ...
            'Input GRDECL is invalid');
   else
      error('buildCornerPtNodes:General:Unexpected', ...
            'Unexpected error %s\n', excpt.message);
   end
end
