function G = cartGrid(celldim, varargin)
%Construct Cartesian grid in physical space.
%
% SYNOPSIS:
%   G = cartGrid(celldim);
%   G = cartGrid(celldim, physdim);
%
% PARAMETERS:
%   celldim - Vector, length 3, specifying number of
%             cells in each coordinate direction.
%   physdim - Vector, length 3, of physical dimensions.
%             OPTIONAL.  Default value == celldim
%             (i.e., each cell has physical dimension 1-by-1-by-1).
%
% RETURNS:
%   G - Grid structure mostly as detailed in grid_structure, though lacking the
%       fields
%         - G.cells.volumes
%         - G.cells.centroids
%
%         - G.faces.areas
%         - G.faces.normals
%         - G.faces.centroids
%
%       These fields may be computed using the function computeGeometry.
%
%       There is, however, an additional field not described in grid_structure:
%
%         - cartDims -- A length 3 vector giving number of cells in each
%                       coordinate direction.  In other words
%
%                               cartDims == celldim .
%
%  New:  G.cellFaces(:,2) contains integers 1-6 corresponding to directions
%        W, E, S, N, T, B.
%
% EXAMPLE: 
%   % Make a 10-by-5-by-2 grid on the unit cube.  
%      nx = 10; ny = 5; nz = 2;
%      G = cartGrid([nx, ny, nz], [1 1 1]);
%   
%   % Plot the grid in 3D-view.
%      f = plotGrid(G); view(3);
%
% SEE ALSO:
%   grid_structure, tensorGrid, computeGeometry.

%{
Copyright 2009 SINTEF ICT, Applied Mathematics.

This file is part of The MATLAB Reservoir Simulation Toolbox (MRST).

MRST is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MRST is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MRST.  If not, see <http://www.gnu.org/licenses/>.
%}

% $Id: cartGrid.m 1956 2009-03-31 11:06:38Z bska $

error(nargchk(1, 2, nargin, 'struct'));

if nargin < 2,
   physdim = celldim;
else
   physdim = varargin{1};
end

dx = repmat(physdim(1) ./ celldim(1), [celldim(1), 1]);
dy = repmat(physdim(2) ./ celldim(2), [celldim(2), 1]);
dz = repmat(physdim(3) ./ celldim(3), [celldim(3), 1]);

G  = tensorGrid(dx, dy, dz);
