function fluid = initSimpleFluid(varargin)
%Initialize incompressible two-phase fluid model.
%
% SYNOPSIS:
%   fluid = initSimpleFluid
%   fluid = initSimpleFluid('pn1', pv1, ...)
%
% PARAMETERS:
%   'pn'/pv - List of 'key'/value pairs defining optional parameters.  The
%             supported options are:
%               - mu  -- Viscosities. Default value = [1 10]     [cP]
%               - rho -- Densities.   Default value = [1000 700] [kg/m^3]
%               - n   -- Exponents.   Default value = [2, 2]
%
% RETURNS:
%   fluid - Fluid data structure representing the current state of the
%           fluids within the reservoir model. Contains scalar fields and
%           function handles that take a structure sol containing a field s
%           (saturation) as argument. sol is normally the reservoir
%           solution structure.
%               -- Scalar fields:
%                    - N      -- Number of phases = 2.
%                    - mu     -- Viscosities.
%                    - rho    -- Densities.
%               -- Function handles:          
%                    - kr     -- Relative permeabilities.  
%                    - dkr    -- Derivative of relative permeabilities. 
%                    - mob    -- Mobilities.
%                    - Lt     -- Total mobility.
%                    - omega  -- Accumulated phase densities weighted by
%                                fractional flow functions. 
%
% NOTE:
%   The fluid model represented by the return argument is the two-phase
%   incompressible counterpart to the fluid model of the Black Oil 'pvt'
%   function.  For generality, the argument to all member functions is
%   assumed to be a structure that contains (at least) a saturation field,
%   's'.  Therefore, to plot the relative permeability, use statements akin
%   to
%
%      s = linspace(0, 1, 20).'; kr = fluid.kr(struct('s', s));
%      plot(s, kr), legend('kr_1', 'kr_2')
%
% SEE ALSO:
%   pvt, initResSol, initWellSol, solveIncompFlow.

%{
Copyright 2009 SINTEF ICT, Applied Mathematics.

This file is part of The MATLAB Reservoir Simulation Toolbox (MRST).

MRST is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MRST is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MRST.  If not, see <http://www.gnu.org/licenses/>.
%}

% $Id: initSimpleFluid.m 2138 2009-05-04 17:09:29Z bska $

opt = struct('mu' ,    [1.0,     10.0], ...
             'rho',    [1000.0, 700.0], ...
             'n'  ,    [2,          2]);
          
opt = merge_options(opt, varargin{:});

n     = opt.n; 
kr    = @(sol) [sol.s(:,1).^n(1), (1-sol.s(:,1)).^n(2)];
dkr   = @(sol) [n(1) .*  sol.s(:,1)   .^(n(1)-1), ...
                n(2) .* (sol.s(:,1)-1).^(n(2)-1)];

fluid = struct('N',     2,              ...
               'kr',    kr,             ...
               'dkr',   dkr,            ...
               'mu',    opt.mu ./ 1000, ...  % cP -> Pa s
               'rho',   opt.rho);

fluid = initFluid(fluid);
