function fluid = initSingleFluid(varargin)
%Initialize incompressible single-phase fluid model.
%
% SYNOPSIS:
%   fluid = initSingleFluid
%   fluid = initSingleFluid('pn1', pv1, ...)
%
% PARAMETERS:
%   'pn'/pv - List of 'key'/value pairs defining optional parameters.  The
%             supported options are:
%               - mu  -- Viscosity. Default value =    1.0 [cP]
%               - rho -- Density.   Default value = 1000.0 [kg/m^3]
%
% RETURNS:
%   fluid - Fluid data structure representing the current state of the
%           fluids within the reservoir model. Contains scalar fields and
%           function handles that take a structure sol containing a field s
%           (saturation) as argument. sol is normally the reservoir
%           solution structure.
%               -- Scalar fields:
%                    - N      -- Number of phases = 1.
%                    - mu     -- Viscosities.
%                    - rho    -- Densities.
%               -- Function handles:          
%                    - kr     -- Relative permeabilities.  
%                    - dkr    -- Derivative of relative permeabilities. 
%                    - mob    -- Mobilities.
%                    - Lt     -- Total mobility.
%                    - omega  -- Accumulated phase densities weighted by
%                                fractional flow functions. 
%
% NOTE:
%   The fluid model represented by the return argument is the single-phase
%   incompressible counterpart to the fluid model of the Black Oil 'pvt'
%   function. For generality, the argument to all member functions is
%   assumed to be a structure that contains (at least) a saturation field.
%   Therefore, to plot the relative permeability, one needs to do as
%   follows
%
%   S.s = linspace(0,1,20)'; plot(S.s,fluid.kr(S));
%
% SEE ALSO:
%   pvt, initResSol, initWellSol, solveIncompFlow.

%{
Copyright 2009 SINTEF ICT, Applied Mathematics.

This file is part of The MATLAB Reservoir Simulation Toolbox (MRST).

MRST is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MRST is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MRST.  If not, see <http://www.gnu.org/licenses/>.
%}

% $Id: initSingleFluid.m 1956 2009-03-31 11:06:38Z bska $

opt = struct('mu' ,    1.0, ...
             'rho', 1000.0);
opt = merge_options(opt, varargin{:});

kr    = @(sol) ones( size(sol.s(:,1)));
dkr   = @(sol) zeros(size(sol.s(:,1)));
mu    = opt.mu;  assert (mu > 0);
rho   = opt.rho;

fluid = struct ('N',     1,           ...
                'kr',    kr,          ...
                'dkr',   dkr,         ...
                'mu',    mu ./ 1000,  ...   % cP -> Pa s
                'rho',   rho);
fluid = initFluid(fluid);
