function src = addSource(src, c, r, varargin)
%Add an explicit source to (new or existing) source object.
%
% SYNOPSIS:
%   src = addSource(src, cells, values)
%   src = addSource(src, cells, values, 'pn1', pv1)
%
% PARAMETERS:
%   src    - Source structure from a prior call to 'addSource' which will
%            be updated on output or an empty array (src==[]) in which case
%            a new source structure is created.
%
%   cells  - Indices in external model for which this source should be
%            applied.
%
%   values - Strength of source.  One scalar value for each cell in
%            'cells'.  Note that the values in 'values' are interpreted as
%            flux rates (typically in units of m^3/Day) rather than as flux
%            density rates (which must be integrated over the cell volumes
%            in order to obtain flux rates).  Specifically, the mimetic
%            pressure solvers do not integrate these values.
%
% OPTIONAL PARAMETERS (supplied in 'key'/value pairs ('pn'/pv ...)):
%   sat    - Fluid composition of injected fluid in injection cells.
%            An n-by-m array of fluid compositions with 'n' being the
%            number of cells in 'cells' and 'm' the number of components in
%            the saturation. For m=3, the columns interpreted as
%              1 <-> Aqua, 2 <-> Liquid, 3 <-> Vapor.
%
%            This field is for the benefit of transport solvers such as
%            'blackoilUpwFE' and will be ignored for production source
%            cells (i.e. when values < 0).
%        
%            Default value: sat = [] (assume single-phase flow).
%
% RETURNS:
%   src - New or updated source structure having the following fields:
%           - cell -- Cells for which explicit sources are provided
%           - rate -- Rates or values of these explicit sources
%           - sat  -- Fluid composition of injected fluids in injection
%                     cells.
%
% EXAMPLE:
%   See simpleSRCandBC.m  
%
% SEE ALSO:
%   addWell, addBC, solveIncompFlow.

%{
Copyright 2009 SINTEF ICT, Applied Mathematics.

This file is part of The MATLAB Reservoir Simulation Toolbox (MRST).

MRST is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MRST is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MRST.  If not, see <http://www.gnu.org/licenses/>.
%}

% $Id: addSource.m 1956 2009-03-31 11:06:38Z bska $

opt = struct('sat', []);
opt = merge_options(opt, varargin{:});
s   = opt.sat;

if isempty(src),
   src = struct('cell', [], 'rate', [], 'sat', []);
end

assert (numel(c) == numel(r));
assert (numel(s)==0 || (size(s,2) >= 1 && numel(r) == size(s,1)));
assert ((size(s,2) == size(src.sat,2)) || (size(src.sat,2) == 0));

src.cell = [src.cell; c(:)];
src.rate = [src.rate; r(:)];
src.sat  = [src.sat ; s];
