function bc = fluxside(bc, G, side, I1, I2, flux, varargin)
%Impose Neumann boundary condition (flux) on global side.
%
% SYNOPSIS:
%   bc = fluxside(bc, G, side, I1, I2, flux)
%   bc = fluxside(bc, G, side, I1, I2, flux, 'pn', pv)
%
% PARAMETERS:
%   bc     - Boundary condition structure as defined by function 'addBC'.
%
%   G      - Grid structure as described by grid_structure.  Currently
%            restricted to grids produced by functions cartGrid and
%            tensorGrid.
%
%   side   - Global side from which to extract face indices.
%            Must be one of {'LEFT', 'RIGHT', 'FRONT', 'BACK', ...
%                            'BOTTOM', 'TOP'}
%
%   I1, I2 - Cell index ranges for local axes one and two, respectively.
%            Must contain *ALL* indices (e.g., I1 = 1:50, I2 = 1:4).
%
%   flux   - Total flux, in units of m^3/s, (scalar) accounted for by
%            faces on side in ranges I1 and I2.
%
%            Note: The 'flux' value is interpreted by the pressure and
%            transport solvers as an injection flux (into the reservoir).
%            Specifically, a positive value is interpreted as an injection
%            flux.  To specify an extraction flux (i.e., flux out of the
%            reservoir), the caller should provide a negative value in
%            'flux'.
%
% OPTIONAL PARAMETERS (supplied in 'key'/value pairs ('pn'/pv ...)):
%   sat    - Fluid composition of fluid injected across inflow faces.
%            An n-by-m array of fluid compositions with 'n' being the
%            number of individual faces specified by (I1,I2) (i.e.,
%            n==NUMEL(I1)*NUMEL(I2)) or one.  If m=3, the columns of 'sat'
%            are interpreted as 1 <-> Aqua, 2 <-> Liquid, 3 <-> Vapor.
%
%            This field is for the benefit of transport solvers such as
%            'blackoilUpwFE' and will be ignored for outflow faces.
%
%            Default value: sat = [] (assume single-phase flow).
%
% RETURNS:
%   bc     - Updated boundary condition structure.
%
% EXAMPLE:
%   See simpleBC.m.
%
% SEE ALSO:
%   pside, addBC, solveIncompFlow, grid_structure.

%{
Copyright 2009 SINTEF ICT, Applied Mathematics.

This file is part of The MATLAB Reservoir Simulation Toolbox (MRST).

MRST is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MRST is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MRST.  If not, see <http://www.gnu.org/licenses/>.
%}

% $Id: fluxside.m 1956 2009-03-31 11:06:38Z bska $

error(nargchk(6, 8, nargin, 'struct'))      

opt = struct('sat', []);
opt = merge_options(opt, varargin{:});
sat = opt.sat;

ix = boundaryFaceIndices(G, side, I1, I2);

assert (any(numel(flux) == [1, numel(ix)]));
assert (numel(sat)==0 || any(size(sat,1) == [1, numel(ix)]));

if size(sat,1) == 1, sat = sat(ones([numel(ix), 1]), :); end

a  = G.faces.areas(ix);
sa = sum(a);
bc = addBC(bc, ix, 'flux', (flux / sa) .* a, 'sat', sat);
