function bc = pside(bc, G, side, I1, I2, pressure, varargin)
%Impose Dirichlet boundary condition (pressure) on global side.
%
% SYNOPSIS:
%   bc = pside(bc, G, side, I1, I2, p)
%   bc = pside(bc, G, side, I1, I2, p, 'pn', pv)
%
% PARAMETERS:
%   bc     - boundary condition structure as defined by function 'addBC'.
%
%   G      - Grid structure as described by grid_structure.  Currently
%            restricted to grids produced by functions cartGrid and
%            tensorGrid.
%
%   side   - Global side from which to extract face indices.
%            Must be one of {'LEFT', 'RIGHT', 'FRONT', 'BACK', ...
%                            'BOTTOM', 'TOP'}
%
%   I1, I2 - Cell index ranges for local axes one and two, respectively.
%            Must contain *ALL* indices (e.g., I1 = 1:50, I2 = 1:4).
%
%   p      - Pressure value, in units of Pascal, to be applied to the face.
%            Either a scalar or a vector of NUMEL(I1)*NUMEL(I2) values.
%
% OPTIONAL PARAMETERS (supplied in 'key'/value pairs ('pn'/pv ...)):
%   sat    - Fluid composition of fluid injected across inflow faces.
%            An n-by-m array of fluid compositions with 'n' being the
%            number of individual faces specified by (I1,I2) (i.e.,
%            n==NUMEL(I1)*NUMEL(I2)) or one.  If m=3, the columns of 'sat'
%            are interpreted as 1 <-> Aqua, 2 <-> Liquid, 3 <-> Vapor.
%
%            This field is for the benefit of transport solvers such as
%            'blackoilUpwFE' and will be ignored for outflow faces.
%
%            Default value: sat = [] (assume single-phase flow).
%
% RETURNS:
%   bc     - Updated boundary condition structure.
%
% EXAMPLE:
%   See simpleBC.m, simpleSRCandBC.m. 
%
% SEE ALSO:
%   fluxside, addBC, solveIncompFlow, grid_structure.

%{
Copyright 2009 SINTEF ICT, Applied Mathematics.

This file is part of The MATLAB Reservoir Simulation Toolbox (MRST).

MRST is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MRST is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MRST.  If not, see <http://www.gnu.org/licenses/>.
%}

% $Id: pside.m 1956 2009-03-31 11:06:38Z bska $

error(nargchk(6, 8, nargin, 'struct'))

opt = struct('sat', []);
opt = merge_options(opt, varargin{:});
sat = opt.sat;

ix = boundaryFaceIndices(G, side, I1, I2);

assert (any(numel(pressure) == [1, numel(ix)]));
assert (numel(sat)==0 || any(size(sat,1) == [1, numel(ix)]));

if size(sat,1)     == 1, sat      = sat(ones([numel(ix), 1]), :); end
if numel(pressure) == 1, pressure = pressure(ones([numel(ix), 1])); end

bc = addBC(bc, ix, 'pressure', pressure, 'sat', sat);
