function varargout = plotFaces(G, faces, varargin)
%Plot selection of coloured grid faces to current axes (reversed Z axis).
%
% SYNOPSIS:
%       plotFaces(G, faces)
%       plotFaces(G, faces, 'pn1', pv1, ...)
%       plotFaces(G, faces, colour)
%       plotFaces(G, faces, colour, 'pn1', pv1, ...)
%   h = plotFaces(...)
%
% PARAMETERS:
%   G       - Grid data structure.
%
%   faces   - Vector of face indices.  The graphical output of 'plotFaces'
%             will be restricted to the subset of grid faces from 'G'
%             represented by 'faces'.
%
%   colour  - Colour data specification.  Either a MATLAB 'ColorSpec'
%             (i.e., an RGB triplet (1-by-3 row vector) or a short or long
%             colour name such as 'r' or 'cyan'), or a PATCH
%             'FaceVertexCData' table suiteable for either indexed or
%             'true-colour' face colouring.  This data *MUST* be an m-by-1
%             column vector or an m-by-3 matrix.  We assume the following
%             conventions for the size of the colour data:
%
%                - SIZE(colour,1) == NUMEL(faces)
%                  One (constant) indexed colour for each face in 'faces'.
%                  This option supports 'flat' face shading only.
%
%                - SIZE(colour,1) == G.nodes.num
%                  One (constant) indexed colour for each node in 'faces'.
%                  This option must be chosen in order to support
%                  interpolated face shading.
%
%             OPTIONAL.  Default value: colour = 'y' (shading flat).
%
%   'pn'/pv - List of other property name/value pairs.  OPTIONAL.
%             This list will be passed directly on to function PATCH
%             meaning all properties supported by PATCH are valid.
%
% RETURNS:
%   h - Handle to resulting PATCH object.  The patch object is added to the
%       current AXES object.
%
% NOTES:
%   Function 'plotFaces' is implemented directly in terms of the low-level
%   function PATCH.  If a separate axes is needed for the graphical output,
%   callers should employ function newplot prior to calling 'plotFaces'.
%
% EXAMPLE:
%   % Plot grid with boundary faces on left side in red color:   
%   G     = cartGrid([5, 5, 2]);
%   faces = boundaryFaceIndices(G, 'LEFT', 1:5, 1:5);
%   plotGrid( G, 'faceColor', 'none'); view(3);
%   plotFaces(G, faces, 'r');
%     
% SEE ALSO:
%   plotCellData, plotGrid, newplot, patch, shading.

%{
Copyright 2009 SINTEF ICT, Applied Mathematics.

This file is part of The MATLAB Reservoir Simulation Toolbox (MRST).

MRST is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MRST is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MRST.  If not, see <http://www.gnu.org/licenses/>.
%}

% $Id: plotFaces.m 1956 2009-03-31 11:06:38Z bska $

   plotOutline = false;
   if numel(varargin)>0,

      i=1+isnumeric(varargin{1});
      v = find(strcmpi({varargin{i:2:end}}, 'outline'));
      if ~isempty(v), 
         % Find argument following last 'outline'
         plotOutline = varargin{i+2*v(end)-1};
         
         % remove pairs 'outline'/value from varargin
         varargin([i+2*v-2, i+2*v-1]) = [];
      end
   end
         
   


   if isempty(faces), return, end

   if ~isnumeric(faces),
      error(msgid('FaceList:NonNumeric'), ...
            'Face list ''faces'' is not numeric.')
   end
   
   if mod(numel(varargin), 2) == 0,
      colour   = 'yellow';
   else
      colour   = varargin{1};
      varargin = { varargin{2 : end} };
   end

   % Extract face topology for subset of grid faces and the coordinates for
   % the actual vertices ('verts') present in this topology.
   %
   [f, verts] = get_face_topo (G, faces);
   v          = G.nodes.coords(verts, :);

   % Massage colour data into form suitable for 'FaceVertexCData' property.
   %
   % From here, we assume that 'colour' is an m-by-1 (for indexed
   % colouring) or m-by-3 (for 'true colour' colouring) array.
   % Furthermore, 'm' is assumed to be either 1 (meaning all 'faces' should
   % be coloured using a single colour), NUMEL(faces) when the individual
   % faces are coloured separately, or G.nodes.num for interpolated face
   % colouring.
   %
   if ischar(colour), colour = get_rgb(colour); end

   if any(size(colour,1) == [1, numel(faces)]),
      fc     = 'flat';
   elseif size(colour,1) == G.nodes.num,
      colour = colour(verts,:);
      fc     = 'interp';
   end

   % Build final patch for graphical output (Note: added to GCA).
   h = patch('Faces'          , f      , 'Vertices' , v , ...
             'FaceVertexCData', colour , 'FaceColor', fc, varargin{:});

   set(get(h, 'Parent'), 'ZDir', 'reverse')

   if plotOutline,
      pts = findFaceOutline(G, faces);
      hold on, plot3(pts(:,1), pts(:,2), pts(:,3), 'k');hold off;
   end
   
   if nargout > 0, varargout{1} = h; end
   
   
end

%--------------------------------------------------------------------------
% Private helpers follow.
%--------------------------------------------------------------------------

function [f, present] = get_face_topo(G, faces)
   eIX = cumsum([0; double(G.faces.numNodes)]);
   nn  = double(G.faces.numNodes(faces));
   fn  = double(G.faceNodes(mcolon(eIX(faces) + 1, eIX(faces + 1)), 1));

   m   = numel(faces);
   n   = max(nn);
   f   = nan([n, m]);

   % Extract only those nodes/vertices actually present in the subset of
   % grid faces represented by 'faces'.  Create local numbering for these
   % vertices.
   %
   present           = false([G.nodes.num, 1]);
   present(fn)       = true;

   node_num          = zeros([G.nodes.num, 1]);
   node_num(present) = 1 : sum(double(present));

   off = reshape((0 : m - 1) .* n, [], 1);

   f(mcolon(off + 1, off + nn)) = node_num(fn);

   % PATCH requires that the 'Faces' property be a matrix of size
   % (number of faces)-by-(number of vertices).
   %
   f = f .';
end

%--------------------------------------------------------------------------

function rgb = get_rgb(colour)
   switch lower(colour),
      case {'y', 'yellow' }, rgb = [1, 1, 0];
      case {'m', 'magenta'}, rgb = [1, 0, 1];
      case {'c', 'cyan'   }, rgb = [0, 1, 1];
      case {'r', 'red'    }, rgb = [1, 0, 0];
      case {'g', 'green'  }, rgb = [0, 1, 0];
      case {'b', 'blue'   }, rgb = [0, 0, 1];
      case {'w', 'white'  }, rgb = [1, 1, 1];
      case {'k', 'black'  }, rgb = [0, 0, 0];
      otherwise            , rgb = [0, 0, 1]; % Unknown colour -> 'blue'.
   end
end



function pts = findFaceOutline(g, faces)
   if numel(faces)==0, pts = zeros(0,3);return;end
   
   cellNo = rldecode(1:g.cells.num, double(g.cells.numFaces), 2) .';
   sgn    = 2*(cellNo == g.faces.neighbors(g.cellFaces(:,1), 1))-1;
   faceNo = rldecode(1:g.faces.num, double(g.faces.numNodes), 2) .';

   fi     = false(g.faces.num, 1); fi(faces) = true;
   fn     = g.faceNodes(fi(faceNo));

   nodeflag  = false(g.nodes.num, 1);
   nodeflag(fn)=true;

   fe = faceEdges(g);
   fe(sgn(faceNo)<0, :) = fe(sgn(faceNo)<0, [2,1]);

   fe = fe (fi(faceNo),:);
   nodes = double(fe(all(nodeflag(fe), 2), :));
   
   if numel(nodes)>0,
      [nodes,n]=rlencode(sortrows(sort(nodes, 2)));
      nodes(n>1,:)=[];
   end

   pts = nan(size(nodes, 1)*3, 3);
   pts (1:3:end,:) = g.nodes.coords(nodes(:,1),:);
   pts (2:3:end,:) = g.nodes.coords(nodes(:,2),:);
end

function fe = faceEdges(g)
   fe = [g.faceNodes, g.faceNodes([2:end,1])];
   p = cumsum([0; double(g.faces.numNodes(1:end-1))]);
   fe (p+double(g.faces.numNodes),2) = fe (p+1,1);
end
