function varargout = plotGrid(G, varargin)
%Plot exterior grid faces to current axes (reversed Z axis).
%
% SYNOPSIS:
%       plotGrid(G)
%       plotGrid(G, 'pn1', pv1, ...)
%       plotGrid(G, cells)
%       plotGrid(G, cells, 'pn1', pv1, ...)
%   h = plotGrid(...)
%
% PARAMETERS:
%   G       - Grid data structure.
%
%   cells   - Vector of cell indices defining sub grid.  The graphical
%             output of function 'plotGrid' will be restricted to the
%             subset of cells from 'G' represented by 'cells'.
%
%             If empty or unspecified, function 'plotCellData' will behave
%             as if the user defined
%
%                 cells = 1 : G.cells.num
%
%             meaning graphical output will be produced for all cells in
%             the grid model 'G'.
%
%   'pn'/pv - List of property names/property values.  OPTIONAL.
%             This list will be passed directly on to function PATCH
%             meaning all properties supported by PATCH are valid.
%
% RETURNS:
%   h  - Handle to resulting patch object.  The patch object is added
%        directly to the current AXES object (GCA).
%        OPTIONAL.  Only returned if specifically requested.
%
% NOTES:
%   Function 'plotGrid' is implemented directly in terms of the low-level
%   function PATCH.  If a separate axes is needed for the graphical output,
%   callers should employ function newplot prior to calling 'plotGrid'.
%
% EXAMPLE:    
%   G = cartGrid([10, 10, 5]);
%
%   % 1) Plot grid with yellow color on faces (default):
%     figure; plotGrid(G); view(3);
%
%   % 2) Plot grid with no color on faces (transparent faces):  
%     figure; plotGrid(G, 'faceColor', 'none'); view(3);
%
% SEE ALSO:
%   plotGridT, plotCellData, plotFaces, patch, newplot.

%{
Copyright 2009 SINTEF ICT, Applied Mathematics.

This file is part of The MATLAB Reservoir Simulation Toolbox (MRST).

MRST is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MRST is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MRST.  If not, see <http://www.gnu.org/licenses/>.
%}

% $Id: plotGrid.m 1956 2009-03-31 11:06:38Z bska $

if numel(G) > 1,
   error(msgid('Grid:MultiComponent'), ...
         'Cannot plot more than one grid at a time.');
end

% Default to showing the entire grid model.
%
cells = (1 : G.cells.num) .';

if mod(numel(varargin), 2) == 1,
   % Caller requested graphical output from a particular subset of the grid
   % cells.  Honour request, but treat an empty 'cells' argument as if the
   % caller requested the default behaviour (i.e., cells = 1:G.cells.num).
   %
   if isnumeric(varargin{1}) && ~isempty(varargin{1}),
      cells = varargin{1};
   end

   % Strip 'cells' argument off of remaining input argument list.
   %
   varargin = { varargin{2 : end} };
end

% Assert that remaining arguments at least appear to be 'name'/value pairs
% intended for PATCH.
%
assert (all(cellfun(@ischar, { varargin{1 : 2 : end} })));

f = boundaryFaces(G, cells);
h = plotFaces(G, f, 'EdgeColor', 'k', 'FaceColor', 'y', varargin{:});

if nargout > 0, varargout{1} = h; end
