function [ff, gg, hh, dF, dC] = computePressureRHS(g, omega, bc, src)
%Compute right-hand side contributions to pressure linear system.
%
% SYNOPSIS:
%   [f, g, h, dF, dC] = computePressureRHS(G, omega, bc, src)
%
% PARAMETERS:
%   G     - Grid data structure.
%
%   omega - Accumulated phase densities \rho_i weighted by fractional flow
%           functions f_i -- i.e., omega = \sum_i \rho_i f_i.  One scalar
%           value for each cell in the discretised reservoir model, G.
%
%   bc    - Boundary condition structure as defined by function 'addBC'.
%           This structure accounts for all external boundary conditions to
%           the reservoir flow.  May be empty (i.e., bc = struct([])) which
%           is interpreted as all external no-flow (homogeneous Neumann)
%           conditions.
%
%   src   - Explicit source contributions as defined by function
%           'addSource'.  May be empty (i.e., src = struct([])) which is
%           interpreted as a reservoir model without explicit sources.
%
% RETURNS:
%   f, g, h - Direct (block) right hand side contributions as expected by
%             the linear system solvers such as 'schurComplementSymm'.
%
%   dF, dC  - Packed Dirichlet/pressure condition structure.  The faces in
%             'dF' and the values in 'dC'.  May be used to eliminate known
%             face pressures from the linear system before calling a system
%             solver (e.g., 'schurComplementSymm').
%
% SEE ALSO:
%   addBC, addSource, pvt, computeMimeticIP, schurComplement.

%{
Copyright 2009 SINTEF ICT, Applied Mathematics.

This file is part of The MATLAB Reservoir Simulation Toolbox (MRST).

MRST is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MRST is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MRST.  If not, see <http://www.gnu.org/licenses/>.
%}

% $Id: computePressureRHS.m 2138 2009-05-04 17:09:29Z bska $

   ff = grav_pressure(g, omega);
   gg = zeros([g.cells.num, 1]);
   hh = zeros([g.faces.num, 1]);

   if ~isempty(src),
      assert (max([src.cell]) <= g.cells.num);

      ss = accumarray(src.cell, src.rate);
      ii = accumarray(src.cell, 1);
      gg(ii > 0) = gg(ii > 0) + ss(ii > 0);
   end

   dF = false([g.faces.num, 1]);
   dC = [];

   if ~isempty(bc),
      assert (max([bc.face]) <= g.faces.num);
      assert (all(accumarray(bc.face, 1, [g.faces.num, 1]) <= 1));

      isDir = strcmpi('pressure', bc.type);
      isNeu = strcmpi('flux',     bc.type);

      [face, ix] = sort(bc.face(isDir));
      dF(face)   = true;

      % Add Dirichlet/pressure conditions to RHS.
      dC = bc.value(isDir);
      dC = dC(ix); % Order according to increasing face index.
      ff(dF(g.cellFaces(:,1))) = ff(dF(g.cellFaces(:,1))) - dC;

      % Add Neumann/flux conditions to RHS.
      % Negative sign because value is INJECTION flux.
      hh(bc.face(isNeu)) = - bc.value(isNeu);
   end

   assert (~any(dC < 0));  % Pressure conditions should always be non-neg.
end

%--------------------------------------------------------------------------

function ff = grav_pressure(g, omega)
   g_vec = gravity();   % Must be a 1-by-3 row vector for subsequent code.
   dim   = size(g.nodes.coords,2);

   assert (1 < dim && dim < 4);
   assert (all(size(g_vec) == [1,3]));

   if norm(g_vec(1 : dim)) > 0,
      cellno = rldecode(1 : g.cells.num, double(g.cells.numFaces), 2) .';
      cvec   = g.faces.centroids(g.cellFaces(:,1), :) - ...
               g.cells.centroids(cellno          , :);
      ff     = omega(cellno) .* (cvec * g_vec(1:dim).');
   else
      ff     = zeros([size(g.cellFaces,1), 1]);
   end
end
