function wellSol = packageWellSol(flux, press, fW, hW)
%Convert well fluxes and pressures to structure form.
%
% SYNOPSIS:
%   wellSol = packageWellSol(flux, press, fW, hW)
%
% PARAMETERS:
%   flux   - Vector of well fluxes.  One scalar value for each well
%            *perforation* in the model.  Assumed to be stored as
%
%              flux = [wf_11, wf_12, ..., wf_1n1, ...
%                      wf_21, wf_22, ..., wf_2n2, ...
%                      ...
%                      wf_k1, wf_k2, ..., wf_knk]
%
%            with 'k' being the number of wells, 'wf_ij' being the well
%            flux in perforation 'j' of well 'i', and 'ni' being the total
%            number of perforations in well 'i'.
%
%   press  - Vector of well pressures.  One scalar (non-negative) value for
%            each well in the model.
%
%   fW, hW - Output arrays 'fW' and 'hW' respectively from function
%            'unpackWellSystemComponents'.
%
% RETURNS:
%   wellSol - One-dimensional well solution structure array, one element
%             for each well in the model.  Each element, 'wellSol(i)', is a
%             structure having the following fields:
%                - flux     -- All perforation fluxes for well 'i'.
%                - pressure -- The well pressure in well 'i'.
%
% SEE ALSO:
%   unpackWellSystemComponents.

%{
Copyright 2009 SINTEF ICT, Applied Mathematics.

This file is part of The MATLAB Reservoir Simulation Toolbox (MRST).

MRST is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MRST is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MRST.  If not, see <http://www.gnu.org/licenses/>.
%}

% $Id: packageWellSol.m 1956 2009-03-31 11:06:38Z bska $

   nW = sum(cellfun(@(x) numel(x) > 0, fW));         % Number of wells.

   ix_f = cumsum([0, cellfun(@numel, fW)]);
   ix_p = cumsum([0, cellfun(@numel, hW)]);

   wellSol = repmat(struct('flux', [], 'pressure', []), [1, nW]);
   for w = 1 : nW,
      wellSol(w).flux     = -flux (ix_f(w) + 1 : ix_f(w + 1));
      wellSol(w).pressure =  press(ix_p(w) + 1 : ix_p(w + 1));
   end
end
