function blockIx = partitionUI(G, coarseDim)
%Partition grid uniformly in logical space.
%
% SYNOPSIS:
%   blockIx = partitionUI(G, coarseDim)
%
% DESCRIPTION:
%   Partitions a corner point grid in grid_structure format (relatively)
%   uniformly in index (logical) space.  Each coarse block is
%   comprised of roughly the same number of cells from the fine grid.
%
% PARAMETERS:
%   G         - grid_structure structure having either a valid field
%               'G.cells.ijkMap' or both valid fields 'G.cartDims' and
%               'G.cells.indexMap'.
%
%   coarseDim - Number of coarse blocks in each physical direction.
%               Assumed to be a LENGTH 3 vector of integers.
%
% RETURNS:
%   blockIx   - A G.cells.num-by-1 vector mapping cells to coarse blocks.
%
% SEE ALSO:
%   processPartition, partitionLayers.

%{
Copyright 2009 SINTEF ICT, Applied Mathematics.

This file is part of The MATLAB Reservoir Simulation Toolbox (MRST).

MRST is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MRST is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MRST.  If not, see <http://www.gnu.org/licenses/>.
%}

% $Id: partitionUI.m 1956 2009-03-31 11:06:38Z bska $

if ~grid_ok(G),
   error('partitionUI:grid_structure:NOIJK', ...
         'Grid is not a valid grid_structure structure');
end

if numel(coarseDim) ~= 3,
   error('partitionUI:coarseDim:Invalid', ...
         '`coarseDim'' is not valid coarse dimension spec');
end

if isfield(G.cells, 'ijkMap');
   ijk = double(G.cells.ijkMap);
   M   = double(max(ijk));
else
   [ijk{1:3}] = ind2sub(G.cartDims, G.cells.indexMap);
   ijk = double([ijk{:}]);
   M   = double(G.cartDims);
end

blockIx = zeros([G.cells.num, 1]);
for d = 3 : -1 : 1,
   B = coarseDim(d);
   blockIx = lbLinDist(ijk(:,d) - 1, M(d), B) + B*blockIx;
end
blockIx = blockIx + 1;  % Map block numbers to (1..PROD(coarseDim))


function f = lbLinDist(f, M, B)
% lbLinDist -- Load-balanced linear distribution
% See Eric F. Van de Velde, Concurrent Scientific Computing,
% 1994, Springer Verlag, p. 54 (Sect. 2.3) for details.
%
% Maps index set (0..M-1) to blocks (0..B-1).

L = floor(M ./ B);  % Tentative number of cells per coarse block.
R = mod(M, B);      % Additional cells not previously accounted for.
f = max(floor(f ./ (L + 1)), floor((f - R) ./ L));


function bool = grid_ok(G)
bool = ~isempty(G)          && isstruct(G)       && ...
        isfield(G, 'cells') &&                      ...
       ~isempty(G.cells)    && isstruct(G.cells);
bool = bool && ((isfield(G.cells, 'ijkMap') && ...
                 size(G.cells.ijkMap,1) == G.cells.num) || ...
                (isfield(G, 'cartDims') && isfield(G.cells, 'indexMap')));
