function [xr, xw] = solveIncompFlowMS(xr, xw, G, CG, p, S, CS, fluid, varargin)
%Solve coarse (multiscale) pressure system.
%
% SYNOPSIS:
%  [xr, xw] = solveIncompFlowMS(xr, xw, G, CG, p, S, CS, fluid)
%  [xr, xw] = solveIncompFlowMS(xr, xw, G, CG, p, S, CS, fluid, ...
%                               'pn1', pv1, ...)
%
% DESCRIPTION:
%   This function assembles and solves a (block) system of linear equations
%   defining interface fluxes and cell and interface pressures at the next
%   time step in a sequential splitting scheme for the reservoir simulation
%   problem defined by Darcy's law and a given set of external influences
%   (wells, sources, and boundary conditions).
%
% REQUIRED PARAMETERS:
%   xr, xw - Reservoir and well solution structures either properly
%            initialized from functions 'initResSol' and 'initWellSol'
%            respectively, or the results from a previous call to function
%            'solveIncompFlow' and, possibly, a transport solver such as
%            function 'implitcitTransport'.
%
%   G, CG, p -
%            Grid, coarse grid, and cell-to-block partition vector,
%            respectively.
%
%   S, CS  - Linear system structure on fine grid (S) and coarse grid
%            (CS) as defined by functions 'computeMimeticIP' and
%            'generateCoarseSystem', respectively.
%
%   fluid  - Fluid object as defined by function 'initSimpleFluid'.
%
% OPTIONAL PARAMETERS (supplied in 'key'/value pairs ('pn'/pv ...)):
%   wells  - well linear system structure (W) as defined by functions
%            'addWell', 'assembleWellSystem', and
%            'assembleCoarseWellSystem'.  An empty well linear system
%            structure (i.e., W=struct([])) is supported.  The resulting
%            model is interpreted as not containing any well driving
%            forces.
%
%   bc     - Boundary condition structure as defined by function 'addBC'.
%            This structure accounts for all external boundary conditions
%            to the reservoir flow.  May be empty (i.e., bc = struct([]))
%            which is interpreted as all external no-flow (homogeneous
%            Neumann) conditions.
%
%   src    - Explicit source contributions as defined by function
%            'addSource'.  May be empty (i.e., src = struct([])) which is
%            interpreted as a reservoir model without explicit sources.
%
%   Solver - Which solver mode function 'solveIncompFlowMS' should employ
%            in assembling and solving the block system of linear
%            equations.
%            String.  Default value: Solver = 'hybrid'.
%
%            Supported values are:
%              - 'hybrid' --
%                 Assemble and solve hybrid system for interface pressures.
%                 System is eventually solved by Schur complement reduction
%                 and back substitution.  The hybrid solver can not be
%                 employed if any of the basis functions in the coarse
%                 system were generated with a positive overlap.
%
%               - 'mixed' --
%                  Assemble and solve a mixed system for cell pressures and
%                  interface fluxes.
%
%   LinSolve -
%            Handle to linear system solver software to which the fully
%            assembled system of linear equations will be passed.  Assumed
%            to support the syntax
%
%                        x = LinSolve(A, b)
%
%            in order to solve a system Ax=b of linear equations.
%            Default value: LinSolve = @mldivide (backslash).
%
%   MatrixOutput -
%            Whether or not to return the final system matrix 'A' to the
%            caller of function 'solveIncompFlowMS'.
%            Logical.  Default value: MatrixOutput = FALSE.
%
% RETURNS:
%   xr - Reservoir solution structure with new values for the fields:
%          - cellPressure -- Pressure values for all cells in the
%                            discretised reservoir model, 'G'.
%          - facePressure -- Pressure values for all interfaces in the
%                            discretised reservoir model, 'G'.
%          - cellFlux     -- Outgoing flux across each local interface for
%                            all cells in the model.
%          - faceFlux     -- Flux across global interfaces corresponding to
%                            the rows of 'G.faces.neighbors'.
%          - A            -- System matrix.  Only returned if specifically
%                            requested by setting option 'MatrixOutput'.
%
%   xw - Well solution structure array, one element for each well in the
%        model, with new values for the fields:
%           - flux     -- Perforation fluxes through all perforations for
%                         corresponding well.  The fluxes are interpreted
%                         as injection fluxes, meaning positive values
%                         correspond to injection into reservoir while
%                         negative values mean production/extraction out of
%                         reservoir.
%           - pressure -- Well pressure.
%
% NOTE:
%   If there are no external influences, i.e., if all of the structures
%   'W', 'bc', and 'src' are empty and there are no effects of gravity,
%   then the input values 'xr' and 'xw' are returned unchanged and a
%   warning is printed in the command window.  This warning is printed with
%   message ID
%
%           'solveIncompFlowMS:DrivingForce:Missing'
%
% SEE ALSO:
%   generateCoarseSystem, computeMimeticIP, addBC, addSource, addWell,
%   initSimpleFluid, initResSol, initWellSol, solveIncompFlow.

%{
#COPYRIGHT#
%}

% $Id: solveIncompFlowMS.m 2138 2009-05-04 17:09:29Z bska $

   opt = struct('bc', [], 'src', [], 'wells', [], ...
                'Solver',       'hybrid',         ...
                'LinSolve',     @mldivide,        ...
                'MatrixOutput', false);
   opt = merge_options(opt, varargin{:});

   g_vec   = gravity();
   no_grav = ~(norm(g_vec(1 : size(G.nodes.coords,2))) > 0);
   if all([isempty(opt.bc)   , ...
           isempty(opt.src)  , ...
           isempty(opt.wells), no_grav]),
      warning(id('DrivingForce:Missing'),                      ...
             ['No external driving forces present in model--', ...
              'state remains unchanged.\n']);
    else
       [A, b, dF, dC, Phi, Psi] = build_system(xr, G, CG, p, S, CS, ...
                                               fluid, opt);
       solver   = pick_solver(CG, CS, dF, dC, opt);
       x        = solver(A, b);
       [xr, xw] = pack_solution(xr, xw, G, CG, p, CS, x{1:3}, ...
                                Phi, Psi, opt);

       if opt.MatrixOutput, xr.A = x{4}; end
    end
end

%--------------------------------------------------------------------------
% Helpers follow.
%--------------------------------------------------------------------------


function [A, b, dF, dC, Phi, Psi] = ...
      build_system(xr, G, CG, p, S, CS, fluid, opt)

   [nsub, sub] = get_subfaces(G, CG, CS);
   [I, J]      = coarsening_ops(G, CG, p, CS, sub, nsub);

   mob   = fluid.Lt(xr);     % == total mobility for all cells
   omega = fluid.omega(xr);  % == \sum_i \rho_i f_i for all cells

   % Build system components B, C, D, f, g, and h for final hybrid system
   %
   %    [  B   C   D  ] [  v ]     [ f ]
   %    [  C'  0   0  ] [ -p ]  =  [ g ] .
   %    [  D'  0   0  ] [ cp ]     [ h ]
   %
   % The components are represented as one-dimensional cell arrays as
   %
   %    A = { B, C, D }
   %    b = { f, g, h }
   %
   % both of which are assembled separately from reservoir contributions
   % (syscomp_res) and well contributions (syscomp_well).  We assemble the
   % global components by concatenating the well components onto the
   % reservoir components.  This assembly is a result of treating wells as
   % faces/contacts.
   %
   % Specifically, we assemble the components as
   %
   %    A{1} = B =   Br   +    Bw    % A{1}(wr,wr) = ...
   %    A{2} = C = [ Cr   ;    Cw ]  % VERTCAT
   %    A{3} = D = [ Dr, 0; 0, Dw ]  % BLKDIAG
   %
   %    b{1} = f = [ fr   ;    fw ]  % VERTCAT
   %    b{2} = g = [ gr   ;       ]  % No direct contr. to block from well
   %    b{3} = h = [ hr   ;    hw ]  % VERTCAT
   %
   [Ar,               Psi_r, Phi_r] = syscomp_res(G, CG, CS, p, mob, opt);
   [Aw, bw, dFw, dCw, Psi_w, Phi_w] = syscomp_wells(opt.wells, G, p, ...
                                                    mob, fluid.rho);

   Psi    = S.BI * [Psi_r, -Psi_w];
   Phi    =        [Phi_r,  Phi_w];

   [br{1:3}, dFr, dCr] = sys_rhs(G, omega, Psi(:,1:size(Psi_r,2)), ...
                                 I, J, opt.src, opt.bc);

   cellno = rldecode(1 : G.cells.num, double(G.cells.numFaces), 2) .';
   Ar{1}  = [Psi_r, -Psi_w].' * ...
            (spdiags(1 ./ mob(cellno), 0, numel(cellno), numel(cellno)) * Psi);

   A    = cell([1, 3]);              b    = cell([1, 3]);
   A{1} = Ar{1}                   ;  b{1} = vertcat(br{1}, bw{1}{:});
   A{2} = vertcat(Ar{2}, Aw{2}{:});  b{2} =         br{2}           ;
   A{3} = blkdiag(Ar{3}, Aw{3}{:});  b{3} = vertcat(br{3}, bw{2}{:});

   wr          = size(Psi_r,2) + (1 : size(Psi_w,2));
   A{1}(wr,wr) = A{1}(wr,wr) + blkdiag(Aw{1}{:});
   if strcmpi(opt.Solver, 'hybrid'),
      A{1} = invert(A{1}, A{2});
   end

   % Finally, assemble a list of prescribed (contact) pressures, i.e., both
   % known face pressure and known bottom hole pressures.  These will be
   % eliminated from the linear system prior to calling one of the system
   % solvers, and entered directly into the solution component 'cp'.
   %
   dF   = [dFr; dFw];
   dC   = [dCr; dCw];
end

%--------------------------------------------------------------------------

function [A, Bv, Phi] = syscomp_res(g, cg, cs, p, mob, opt)
   % Form hybrid or mixed basis function matrices depending on user
   % requests and basis function suitability (cannot employ hybrid
   % formulation if any basis function were generated with positive
   % overlap).
   %
   if strcmpi(opt.Solver, 'hybrid'),
      if any(cellfun(@(x) x{7} > 0, { cs.basisP{cs.activeFaces} })),
         error('solveIncompFlowMS:Solver:NotApplicable', '%s\n%s',    ...
               'The ''hybrid'' coarse solver is not applicable when', ...
               '    any basis function is generated on an extended domain.');
      end
      [Bv, Phi] = hybrid_basis(g, cg, cs);
   else
      [Bv, Phi] = mixed_basis (g, cg, cs);
   end

   % Update pressure basis function matrix for effects of (modified?)
   % coarse block mobilities.  The average mobility is stored in the sixth
   % element of the input tuples (see evalBasisFunc for details).  We do
   % rely on the basis function average mobility to contain one or two
   % elements in the 'hybrid' case (two when the basis function is
   % supported in two distinct coarse blocks), and a single element in the
   % 'mixed' case.
   %
   % The update is a diagonal matrix containing the entries
   %    mob(block)_0 / mob(block)_now
   %
   mob0 = cellfun(@(x) x{6}, { cs.basisP{cs.activeFaces} }, ...
                  'UniformOutput', false);
   mob  = accumarray(p, mob .* g.cells.volumes, [cg.cells.num, 1]) ./ ...
          accumarray(p,        g.cells.volumes, [cg.cells.num, 1]);
   mob  = vertcat(mob0{:}) ./ mob(cg.cellFaces(cs.activeCellFaces,1));
   Phi  = Phi * spdiags(mob, 0, size(Phi,2), size(Phi,2));

   % Extract the 'C' and 'D' matrices for active coarse faces.
   A    = cell([1, 3]);
   A{2} = cs.C(cs.activeCellFaces,        :      );
   A{3} = cs.D(cs.activeCellFaces, cs.activeFaces);
end

%--------------------------------------------------------------------------

function [A, b, dF, dC, Psi, Phi] = syscomp_wells(w, g, p, mob, rho)
   A = cell([1, 3]);
   b = cell([1, 2]);

   [A{:}, b{:}, Psi, Phi] = unpackWellSystemComponentsMS(w, g, p, mob, rho);

   if isempty(w),
      dF = logical([]);
      dC =         [] ;
   else
      dF = reshape(strcmp({ w.type } .', 'bhp'), [], 1);
      dC = reshape([ w(dF).val ]               , [], 1);
   end
end

%--------------------------------------------------------------------------

function [Bv, Phi] = hybrid_basis(g, cg, cs)
   % Form hybrid basis function matrices B*\Psi (Bv) and \Phi (Phi).
   % Called from syscomp_res (only).

   % 1) Enumerate the (active) cell face unknowns according to the
   %    'cs.activeCellFaces' field.  As the basis functions are stored in
   %    the cs.activeFaces elements of the various 'cs.basis*' fields, we
   %    also create an alias for these coarse faces.
   %
   actB        = cs.activeCellFaces;
   actF        = cs.activeFaces;
   renum       = zeros([size(cg.cellFaces,1), 1]);
   renum(actB) = 1 : numel(actB);

   % 2) Build a block-to-face renumbering engine.  Specifically,
   %      f2hf_c(b,f) = j
   %    means that coarse face 'f' within coarse block 'b' corresponds to
   %    coarse flux unknown (coarse half-face) 'j' (when j > 0).
   %
   [nb, nf] = deal(double(cg.cells.num), double(cg.faces.num));
   f2hf_c   = sparse(double(cg.cellFaces(:,1)), ...
                     double(cg.cellFaces(:,2)), renum, nb, nf);

   % 3) Apply hybrid basis function splitting to each (active) flux basis
   %    function.  This process forms the SPARSE input triplet [i,j,v] from
   %    which the hybrid flux basis matrix Bv is finally formed.
   %
   %    Each element of the cell array { cs.basis{actF} } is a 'V' tuple as
   %    defined by function 'evalBasisFunc'.  We also need to renumber the
   %    block-to-face connections, so pass the static 'f2hf_c' matrix along
   %    with the dynamic 'V' tuples.
   %
   [i,j,v] = cellfun(@(x) split_for_hybrid(x, f2hf_c), ...
                     { cs.basis{actF} }, 'UniformOutput', false);
   Bv = sparse(vertcat(i{:}), vertcat(j{:}), vertcat(v{:}), ...
               size(g.cellFaces,1), numel(actB));

   % 4) Apply hybrid basis function splitting to each (active) pressure
   %    basis function.  This process forms the SPARSE input triplet
   %    [i,j,v] from which the hybrid pressure basis matrix Phi is finally
   %    formed.
   %
   %    Each element of the cell array { cs.basisP{actF} } is a 'P' tuple
   %    as defined by function 'evalBasisFunc'.
   %
   [i,j,v] = cellfun(@(x) split_for_hybrid(x, f2hf_c), ...
                     { cs.basisP{actF} }, 'UniformOutput', false);
   Phi = sparse(vertcat(i{:}), vertcat(j{:}), vertcat(v{:}), ...
                g.cells.num, numel(actB));
end

%--------------------------------------------------------------------------

function [i, j, v] = split_for_hybrid(c, hfno)
   % Split a basis function tuple (either flux or pressure) 'c' into a
   % SPARSE matrix input triplet [i,j,v].  A hybrid basis function is split
   % into one component for each of (at most) two coarse half-faces.  We
   % recall that the input tuple, 'c', is a cell array of the form
   %
   %    1  2  3  4  5  6  7
   %   {i, s, f, b, n, m, o}
   %
   % with symbols as defined in 'evalBasisFunc'.  Items 6 and 7 are ignored
   % in this part of the implementation, while item 1 directly provides the
   % desired output vector 'i' (no further processing required).
   %
   % Our task, then, is to convert item 2 into the output vector 'v', and
   % to build output vector 'j' from scratch.  The output 'v' is the
   % simplest to define.  We simply negate the v-values (item 2)
   % corresponding to the second block (b(2)) if applicable.  If
   % numel(b)==1, then v==c{2} is returned unchanged.  The 'j' vector
   % contains at most two distinct values, each corresponding to a specific
   % coarse half-face identified by a pair (b,f) of coarse blocks and face.

   % 1) Extract 'i' values directly from input tuple.
   i = c{1};

   % 2) Build 'j' values from scratch by
   %    i   - Defining pairs of coarse blocks and face.
   b = reshape(double(c{4}), [], 1);
   f = repmat (double(c{3}), [numel(b), 1]);

   %    ii  - Extracting coarse half-faces identified by these pairs.
   j = reshape(full(hfno(sub2ind(size(hfno), b, f))), [], 1);

   %    iii - Expanding these half-face indices a number of times equal to
   %          the number of entities (item 5) in each of the coarse blocks
   %          identified by b.
   j = rldecode(j, reshape(c{5}, [], 1));

   % 3) Negate b(2) values if numel(b) > 1.  Else somewhat expensive NOP.
   s = [1; -1];
   v = rldecode(s(1 : numel(b)), reshape(c{5}, [], 1)) .* c{2};
end

%--------------------------------------------------------------------------

function [Bv, Phi] = mixed_basis(g, cg, cs)
   % Form mixed basis function matrices B*\Psi (Bv) and \Phi (Phi).  Called
   % from syscomp_res (only).

   % 1) Enumerate the (active) face unknowns according to the
   %    'cs.activeFaces' field.
   %
   actF        = cs.activeFaces;
   renum       = zeros([cg.faces.num, 1]);
   renum(actF) = 1 : numel(actF);

   % 2) Apply mixed basis function expansion to each (active) flux basis
   %    function.  This process forms the SPARSE input triplet [i,j,v] from
   %    which the hybrid flux basis matrix Bv is finally formed.
   %
   %    Each element of the cell array { cs.basis{actF} } is a 'V' tuple as
   %    defined by function 'evalBasisFunc'.  We need to renumber the
   %    coarse face flux unknowns, so pass the face enumeration array along
   %    with the dynamic 'V' tuples.  Finally, as the flow direction across
   %    face 'f' is from block cg.faces.neighbors(f,1) to block
   %    cg.faces.neighbors(f,2), the expansion process needs access to the
   %    block topology in order to determine the direction of the flow.
   %
   [i,j,v] = cellfun(@(x) expand_mixed(x, renum, cg.faces.neighbors), ...
                     { cs.basis{actF} }, 'UniformOutput', false);
   Bv = sparse(vertcat(i{:}), vertcat(j{:}), vertcat(v{:}), ...
               size(g.cellFaces,1), numel(actF));

   % 3) Apply mixed basis function expansion to each (active) pressure
   %    basis function.  This process forms the SPARSE input triplet
   %    [i,j,v] from which the hybrid pressure basis matrix Phi is finally
   %    formed.
   %
   %    Each element of the cell array { cs.basisP{actF} } is a 'P' tuple
   %    as defined by function 'evalBasisFunc'.
   %
   [i,j,v] = cellfun(@(x) expand_mixed(x, renum, cg.faces.neighbors), ...
                     { cs.basisP{actF} }, 'UniformOutput', false);
   Phi = sparse(vertcat(i{:}), vertcat(j{:}), vertcat(v{:}), ...
                g.cells.num, numel(actF));
end

%--------------------------------------------------------------------------

function [i, j, v] = expand_mixed(c, fno, n)
   % Expand a basis function tuple (either flux or pressure) 'c' into a
   % SPARSE matrix input triplet [i,j,v].  The values of a mixed basis
   % function are entered directly into the corresponding column (active
   % coarse face number) of the matrix.  We recall that the input tuple,
   % 'c', is a cell array of the form
   %
   %    1  2  3  4  5  6  7
   %   {i, s, f, b, n, m, o}
   %
   % with symbols as defined in 'evalBasisFunc'.  Items 5-7 are ignored in
   % this part of the implementation, while item 1 directly provides the
   % desired output vector 'i' (no further processing required).
   %
   % Our task, then, is to convert item 2 into the output vector 'v', and
   % to build output vector 'j' from scratch.  The 'j' vector is
   % constructed by simply repeating the active coarse face index a number
   % of times equal to the number of values in 'i' (or 'v').  The output
   % 'v' needs to take into account the flow direction induced by the
   % coarse grid topology.  Specifically, the basis functions are generated
   % on the assumption that the flow across coarse face 'f' is from coarse
   % block n(f,1) to coarse block n(f,2).

   % 1) Extract 'i' values directly from input tuple.
   i = c{1};

   % 2) Repeat active coarse face number (fno(c{3})) a number of times
   %    equal to the number of entries in 'i'.
   j = rldecode(fno(c{3}), numel(i));

   % 3) Reverse flow direction if this block (c{4}(1)) is the *inflow*
   %    block for the basis function across coarse face c{3}.
   s = 2*double(n(c{3},1) == c{4}(1)) - 1;
   v = s .* c{2};
end

%--------------------------------------------------------------------------

function [nsub, sub] = get_subfaces(G, CG, CS)
% get_subfaces -- Extract fine-scale faces constituting all active coarse
%                 faces.
%
% SYNOPSIS:
%   [nsub, sub] = get_subfaces(G, CG, CS)
%
% PARAMETERS:
%   G, CG - Fine and coarse grid discretisation of the reservoir model,
%           respectively.
%
%   CS    - Coarse linear system structure as defined by function
%           'generateCoarseSystem'.
%
% RETURNS:
%   nsub, sub - The return values from function 'subFaces' compacted to
%               only account for active coarse faces (i.e. coarse faces
%               across which there is a non-zero flux basis function and,
%               therefore a possibility of non-zero flux).

   [nsub, sub] = subFaces(G, CG);
   sub_ix      = cumsum([0; nsub]);
   sub         = sub(mcolon(sub_ix(CS.activeFaces) + 1, ...
                            sub_ix(CS.activeFaces + 1)));
   nsub        = nsub(CS.activeFaces);
end

%--------------------------------------------------------------------------

function [I, J] = coarsening_ops(G, CG, p, CS, sub, nsub)
% coarsening_ops -- Compute coarsening operators from fine to coarse model.
%
% SYNOPSIS:
%   [I, J] = coarsening_ops(G, CG, p, CS, sub, nsub)
%
% PARAMETERS:
%   G, CG, p  - Fine and coarse grid discretisation of the reservoir model,
%               as well as cell-to-coarse block partition vector.
%
%   CS        - Coarse linear system structure as defined by function
%               'generateCoarseSystem'.
%
%   sub, nsub - Packed representation of fine-scale faces constituting
%               (active) coarse faces.  Assumed to follow the conventions
%               of the return values from function 'subFaces'.
%
% RETURNS:
%  I - Coarse-to-fine cell scatter operator (such that I.' is the
%      fine-to-coarse block gather (sum) operator).
%
%  J - Coarse-to-fine face scatter operator for active coarse faces such
%      that J.' is the fine-to-coarse face gather (sum) operator.

   I = sparse(1 : G.cells.num, p, 1, G.cells.num, CG.cells.num);
   J = sparse(sub, rldecode(CS.activeFaces(:), nsub), ...
              1, G.faces.num, CG.faces.num);
   J = J(:, CS.activeFaces);
end

%--------------------------------------------------------------------------

function [ff, gg, hh, dF, dC] = sys_rhs(G, omega, Psi, I, J, src, bc)
% sys_rhs -- Evaluate coarse system right hand side contributions.
%
% SYNOPSIS:
%   [f, G, h, dF, dC] = sys_rhs(G, omega, Psi, I, J, src, bc)
%
% PARAMETERS:
%   G     - Grid data structure.
%
%   omega - Accumulated phase densities \rho_i weighted by fractional flow
%           functions f_i -- i.e., omega = \sum_i \rho_i f_i.  One scalar
%           value for each cell in the discretised reservoir model, G.
%
%   Psi   - Flux basis functions from reservoir blocks.
%
%   I, J  - Coarsening operators as defined by function 'coarsening_ops'.
%
%   bc    - Boundary condition structure as defined by function 'addBC'.
%           This structure accounts for all external boundary conditions to
%           the reservoir flow.  May be empty (i.e., bc = struct([])) which
%           is interpreted as all external no-flow (homogeneous Neumann)
%           conditions.
%
%   src   - Explicit source contributions as defined by function
%           'addSource'.  May be empty (i.e., src = struct([])) which is
%           interpreted as a reservoir model without explicit sources.
%
% RETURNS:
%   f, g, h - Coarse-scale direct (block) right hand side contributions as
%             expected by the linear system solvers such as
%             'schurComplement'.
%
%   dF, dC  - Coarse-scale packed Dirichlet/pressure condition structure.
%             The faces in 'dF' and the values in 'dC'.  May be used to
%             eliminate known face pressures from the linear system before
%             calling a system solver (e.g., 'schurComplement').
%
% SEE ALSO:
%   computePressureRHS.

   % Evaluate fine-scale system right hand side contributions.
   %
   [ff, gg, hh, dF, dC] = computePressureRHS(G, omega, bc, src);

   % Accumulate fine-scale boundary conditions to coarse scale.
   %
   ff = Psi.' * ff;
   gg =  I .' * gg;
   hh  = J .' * hh;

   % Account for coarse scale Dirichlet boundary conditions.
   %
   dC2 = zeros(size(dF));  dC2(dF) = dC;
   dF  = J.' * dF;
   dC  = J.' * dC2 ./ dF;
   dF  = logical(dF);
   dC  = dC(dF);
end

%--------------------------------------------------------------------------

function BI = invert(B, C)
   % Pre-allocate storage for sparse input arrays.
   %
   n = sum(sum(C > 0) .^ 2);
   i = zeros([n, 1]);
   j = zeros([n, 1]);
   s = zeros([n, 1]);

   % Compute inverse mass matrix values by inverting local matrices
   % (one dense matrix for each coarse block).
   %
   ix = 0;
   for k = 1 : size(C,2),
      r   = reshape(find(C(:,k)), [], 1);
      nF  = numel(r);
      nF2 = nF^2;

      i(ix + 1 : ix + nF2) = repmat(r  , [nF, 1]);
      j(ix + 1 : ix + nF2) = repmat(r.', [nF, 1]);
      s(ix + 1 : ix + nF2) = inv( B(r,r) );

      ix = ix + nF2;
   end

   % Assemble final inverse mass matrix.
   %
   BI = sparse(i, j, s, size(B,1), size(B,2));
end

%--------------------------------------------------------------------------

function s = id(s)
   s = ['solveIncompFlowMS:', s];
end

%--------------------------------------------------------------------------

function solver = pick_solver(CG, CS, dF, dC, opt)
   regul = ~any(dF);  % Set zero level if no prescribed pressure values.

   switch lower(opt.Solver),
      case 'hybrid', solver = @solve_hybrid;
      case 'mixed' , solver = @solve_mixed ;
      otherwise,
         error(id('SolverMode:NotSupported'), ...
               'Solver mode ''%s'' is not supported.', opt.solver);
   end

   %-----------------------------------------------------------------------
   % Enter prescribed pressures into final solution vector ----------------
   %
   lam     = zeros(size(dF));
   lam(dF) = dC;

   function x = solve_hybrid(A, b)
      A{3} = A{3}(:,~dF);
      b{3} = b{3}(  ~dF);
      [x{1:4}] = schurComplementSymm(A{:}, b{:},          ...
                                     'Regularize', regul, ...
                                     'LinSolve', opt.LinSolve);
      lam(~dF) = x{3};
      x{3}     = lam;
   end

   function x = solve_mixed(A, b)
      nF = neumann_faces   (CG, dF, CS.activeFaces, opt);
      Do = oriented_mapping(CG, CS, opt);

      % Reduce size of system by only taking into account the 'Neumann'
      % faces.
      %
      A{3} = A{3}(:,nF);
      b{3} = b{3}(  nF);

      % We have already formed the 'mixed' B matrix in function
      % 'build_system'.  Thus, pass option 'MixedB' to linear solver.
      %
      [x{1:4}] = mixedSymm(A{:}, b{:}, Do, 'Regularize', regul, ...
                           'MixedB', true, 'LinSolve', opt.LinSolve);

      % Put known contact pressures into solution vector.  Note: These
      % contact pressures are only defined non-trivially on the external
      % boundary of the domain.
      %
      lam(nF) = x{3};
      x{3}    = lam;
   end
end

%--------------------------------------------------------------------------

function [xr, xw] = pack_solution(xr, xw, G, CG, p, CS, flux, pres, ...
                                  lam, Phi, Psi, opt)
   [nsub, sub] = get_subfaces(G, CG, CS);
   actB = CS.activeCellFaces;
   actF = CS.activeFaces;

   %-----------------------------------------------------------------------
   % Package reservoir (coarse) solution structure ------------------------
   %
   switch lower(opt.Solver),
      case 'hybrid',
         nhf      = numel(actB);
         flux_sgn = rldecode([1, -1], [nhf, size(Phi,2) - nhf], 2) .';

         xr.blockFlux = flux(1 : numel(actB));
         fluxW        = flux(numel(actB) + 1 : end);
      case 'mixed',
         nf       = numel(actF);
         flux_sgn = rldecode([1, -1], [nf, size(Phi,2) - nf], 2) .';

         orient       = get_orientation(CG, CS);
         renum        = zeros([CG.faces.num, 1]);
         renum(actF)  = 1 : numel(actF);

         xr.blockFlux = orient .* flux(renum(CG.cellFaces(actB,2)));
         fluxW        = flux(numel(actF) + 1 : end);
   end

   cellNo = rldecode(1 : G.cells.num, double(G.cells.numFaces), 2).';

   xr.blockPressure   = pres;
   xr.cellPressure(:) = pres(p) + Phi * (flux .* flux_sgn);
   xr.facePressure(:) = accumarray(G.cellFaces(:,1), ...
                                   xr.cellPressure(cellNo)) ./ ...
                        accumarray(G.cellFaces(:,1), 1);

   if strcmpi(opt.Solver, 'hybrid'),
      % The 'mixed' solver does not generate reasonable coarse interface
      % pressure on internal coarse faces.  Therefore, only insert coarse
      % interface pressures computed by the solver in the 'hybrid' case.
      %
      xr.facePressure(sub) = rldecode(lam(1 : numel(actF)), nsub);
   end

   xr.cellFlux(:)     = Psi * flux;
   xr.faceFlux(:)     = cellFlux2faceFlux(G, xr.cellFlux);

   %-----------------------------------------------------------------------
   % Recover fine scale well fluxes and pressures. ------------------------
   %
   w    = opt.wells;
   lamW = lam(numel(actF) + 1 : end);  assert (numel(lamW) == numel(w));

   nw = numel(w);
   i  = 0;
   for k = 1 : nw,
      rates = w(k).CS.rates;

      xw(k).flux(:)  = - full(rates * fluxW(i + 1 : i + size(rates,2)));
      xw(k).pressure = lamW(k);

      i = i + size(rates,2);
   end
end

%--------------------------------------------------------------------------

function nF = neumann_faces(CG, dF, actF, opt)
% Determine the 'Neumann faces' (nF) (i.e., the faces for which
% face/contact pressures will be determined in the 'mixed' and 'TPFA'
% cases) of the model (g) according to the following rules:
%
%   - No internal faces are counted amongst the 'Neumann faces'.
%   - An external face is a 'Neumann face' unless there is a prescribed
%     pressure value (i.e., a Dirichlet condition) associated to the face.
%   - Wells controlled by rate-constraints are treated as Neumann contacts.
%

   nF = false(numel(actF),1); % No int.
   % All external active faces ... :
   nF(any(CG.faces.neighbors(actF,:) == 0, 2)) = true;
   nF(dF(1:numel(actF))) = false; % ... except Dirichlet cond.

   if ~isempty(opt.wells),
      % Additionally include all rate-constrained wells.
      nF = [nF; strcmp('rate', { opt.wells.type } .')];
   end
end

%--------------------------------------------------------------------------

function orient = get_orientation(cg, cs)
   c1     = cg.faces.neighbors(cg.cellFaces(:,2), 1);
   orient = 2*double((cg.cellFaces(:,1) == c1)) - 1;
   orient = orient(cs.activeCellFaces);
end

%--------------------------------------------------------------------------

function Do = oriented_mapping(CG, CS, opt)
% 'Do' maps face fluxes to half-face fluxes.  This matrix is used to form
% the reduced, mixed system of linear equtions in linear system solver
% function 'mixedSymm'.

   orient = get_orientation(CG, CS);

   if ~isempty(opt.wells),
      ws = [ opt.wells.CS ];
      dw = diag(vertcat(ws.D));
      ow = ones([size(dw,1), 1]);
   else
      dw = [];
      ow = [];
   end
   actB = CS.activeCellFaces;
   actF = CS.activeFaces;
   nf = numel(orient) + numel(ow);

   Do = spdiags([orient; ow], 0, nf, nf) * blkdiag(CS.D(actB, actF), dw);
end
