function T = computeTrans(G, rock, varargin)
%Compute transmissibilities.
%
% SYNOPSIS:
%   T = computeTrans(G, rock)
%   T = computeTrans(G, rock, 'pn1', pv1, ...)
%
% PARAMETERS:
%   G       - Grid data structure as described by grid_structure.
%
%   rock    - Rock data structure with valid field 'perm'.  The
%             permeability is assumed to be in measured in units of
%             metres squared (m^2).  Use function 'darcy' to convert from
%             (milli)darcies to m^2, e.G., perm = perm * darcy() / 1e3 if
%             the permeability is provided in units of millidarcies.
%
%             The field rock.perm may have ONE column for a scalar
%             permeability in each cell, THREE columns for a diagonal
%             permeability in each cell and SIX columns for a symmetric
%             full tensor permeability.  In the latter case, each cell gets
%             the permeability tensor
%                         _          _
%                        | k1  k2  k3 |
%                 K_i  = | k2  k4  k5 |
%                        | k3  k5  k6 |
%                         -          -
%
%   'pn'/pv - List of 'key'/value pairs defining optional parameters.  The
%             supported options are:
%               - Verbose -- Whether or not to emit informational messages
%                            throughout the computational process.
%                            Default value = false.
%
% RETURNS:
%   T - half-transmissibilities for each local face of each grid cell
%       in the grid.  The number of half-transmissibilities equal the
%       number of rows in G.cellFaces.
%
% COMMENTS:
%   PLEASE NOTE: Face normals have length equal to face areas.

%{
Copyright 2009 SINTEF ICT, Applied Mathematics.

This file is part of The MATLAB Reservoir Simulation Toolbox (MRST).

MRST is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MRST is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MRST.  If not, see <http://www.gnu.org/licenses/>.
%}

% $Id: computeTrans.m 2138 2009-05-04 17:09:29Z bska $

opt = struct('verbose', false);
opt = merge_options(opt, varargin{:});

dispif(opt.verbose, 'Computing one-sided transmissibilities...\t');
ticif(opt.verbose);

cellNo = rldecode(1:G.cells.num, double(G.cells.numFaces), 2)';
sgn    = 2*(cellNo == G.faces.neighbors(G.cellFaces(:,1), 1)) -1;
N      = bsxfun(@times, sgn, G.faces.normals(G.cellFaces(:,1),:));
C      = G.faces.centroids(G.cellFaces(:,1), :) - ...
         G.cells.centroids(cellNo, :);
C2     = sum(C.*C, 2);

% Scalar permeability
if size(rock.perm, 2) == 1,
   T = rock.perm(cellNo).*sum (C.*N , 2)./C2;

% Diagonal permeability
elseif size(rock.perm, 2) == 3,
   T = sum(C.*rock.perm(cellNo, :).*C, 2)./C2    .*sum (C.*N , 2)./C2;

% Symmetric permeability
else
   assert(size(rock.perm, 2) == 6);
   i  = [1, 1, 1, 2, 2, 2, 3, 3, 3];
   j  = [1, 2, 3, 2, 4, 5, 3, 5, 6];
   k  = [1, 2, 3, 1, 2, 3, 1, 2, 3];
   T  = sum (C(:, i).*rock.perm(cellNo, j) ...
             .* C(:, k), 2) ./ C2 .* sum (C.*N, 2)./C2;
end
