function [gravmat, flux, q, pv, dt] = initTransport(G, resSol, wellSol, ...
                                                    rock, fluid, varargin)
%Compute input to transport solver from flow calculation results.
%If specified, also determine timestep for use in explicit solver.
%
% SYNOPSIS:
%   [gravmat, ...
%    flux, q, pv, dt] = initTransport(G, resSol, wellSol, rock, fluid)
%   [gravmat, ...
%    flux, q, pv, dt] = initTransport(G, resSol, wellSol,
%                                     rock, fluid, 'pn1', pv1, ...)
%
% PARAMETERS:
%   G       - Grid data structure discretising the reservoir model.
%
%   resSol  - Reservoir solution structure.  Pressures are assumed to be
%             measured in units of Pascal while fluxes are assumed to be
%             measured in units of m^3/s.
%
%   wellSol - Well solution structure.  Pressures are assumed to be
%             measured in units of Pascal while fluxes are assumed to be
%             measured in units of m^3/s.
%
%   rock    - Rock data structure.  Must contain valid fields 'rock.perm',
%             with permeability measured in units of m^2, and 'rock.poro'.
%
%   fluid   - Data structure describing the fluids in the problem.
%
%   'pn'/pv - List of 'key'/value pairs defining optional parameters.  The
%             supported options are:
%
%               - wells  -- Well structure as defined by functions
%                           'addWell' and 'assembleWellSystem'.  May be
%                           empty (i.e., W = struct([])) which is
%                           interpreted as a model without any wells.
%
%               - bc  -- Boundary condtion structure as defined by function
%                        'addBC'.  This structure accounts for all external
%                        boundary contributions to the reservoir flow.
%                        Default value: bc = [] meaning all external
%                        no-flow (homogeneous Neumann) conditions.
%
%               - src -- Explicit source contributions as defined by
%                        function 'addSource'.
%                        Default value: src = [] meaning no explicit
%                        sources exist in the model.
%
%               - OnlyGrav  --  Whether or not to only compute gravity
%                               contribution (for use in gravity
%                               splitting).  Default value: false.
%
%               - ComputeDt --  Whether or not to compute timestep.
%                               Default value: true.
%
%               - max_dt    --  Maximal timestep (in units of seconds).
%                               Default value = 100 * 86400 (100 days).
%
%               - dt_factor --  Factor to multiply the computed dt by.
%                               Default value = 0.5.
%                               Ignored if ComputeDt = false.
%
% RETURNS:
%   flux    - In-flow connection matrix suitable for passing to transport
%             solver.  Content depends on presence or absence of gravity
%             effects:
%
%               - NORM(gravity()) == 0:
%                 flux contains only inflow,
%                 size(flux) = (G.cells.num, G.cells.num)
%               - NORM(gravity()) >  0:
%                 flux contains flow over all non-boundary faces,
%                 size(flux) = (G.cells.num, (G.faces.num-#bndFaces))
%
%             The fluxes are measured in units of m^3/s.
%
%   gravmat - Matrix with gravity contribution for each face.
%             Size G.cells.num-by-(G.faces.num-#bndFaces)).  The gravity
%             flux contributions, if present, are measured in units of
%             m^3/s.
%
%   q       - Aggregate source term suitable for passing to transport
%             solver.  The source term is measured in units of m^3/s.
%
%   pv      - Vector of size G.cells.num-by-1 of cell pore volumes.
%
%   dt      - Timestep (in units of seconds) for use in explicit solver.
%
% SEE ALSO:
%   explicitTransport, implicitTransport, solveIncompFlow, gravity,
%   twophaseUpwFE, twophaseUpwFEGrav, twophaseUpwBE, twophaseUpwBEGrav.

%{
Copyright 2009 SINTEF ICT, Applied Mathematics.

This file is part of The MATLAB Reservoir Simulation Toolbox (MRST).

MRST is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MRST is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MRST.  If not, see <http://www.gnu.org/licenses/>.
%}

% $Id: initTransport.m 2138 2009-05-04 17:09:29Z bska $

   opt = struct('OnlyGrav' , false,       ...
                'ComputeDt', true,        ...
                'max_dt'   , 100 * day(), ...
                'dt_factor', 0.5,         ...
                'wells'    , [],          ...
                'src'      , [],          ...
                'bc'       , []);

   opt = merge_options(opt, varargin{:});
   assert (opt.dt_factor > 0 && opt.max_dt > 0);

   %-----------------------------------------------------------------------
   %% Assemble flux and gravity matrices ----------------------------------
   %
   is_int = all(double(G.faces.neighbors) > 0, 2);
   g_vec  = gravity();

   if norm(g_vec(1 : size(G.nodes.coords,2))) > 0,
      [gravmat, flux] = matrices(G, rock, fluid, resSol, ...
                                 opt.OnlyGrav, is_int);
   else
      [gravmat, flux] = matrices_nograv(G, resSol, is_int);
   end

   %-----------------------------------------------------------------------
   %% External sources/sinks (e.g. wells and BC's) ------------------------
   %
   qi = [];  % Cells to which sources are connected
   qs = [];  % Actual strength of source term (in m^3/s).

   if ~isempty(opt.wells),
      [i, s] = contrib_wells(opt.wells, wellSol);
      qi = [qi; i];
      qs = [qs; s];
   end

   if ~isempty(opt.src), assert (~isempty(opt.src.sat))
      [i, s] = contrib_src(opt.src);
      qi = [qi; i];
      qs = [qs; s];
   end

   if ~isempty(opt.bc), assert (~isempty(opt.bc.sat))
      [i, s] = contrib_bc(G, resSol, opt.bc, is_int);
      qi = [qi; i];
      qs = [qs; s];
   end

   %-----------------------------------------------------------------------
   %% Assemble final phase flux and source contributions in SPARSE format -
   %
   q  = sparse(qi, 1, qs, G.cells.num, 1);
   pv = poreVolume(G, rock);

   %-----------------------------------------------------------------------
   %% Compute CFL-number/dt -----------------------------------------------
   %
   % CFL-condition:
   %    dt <= min(pv ./ (flux_in * max{f'(s)}_{0<=s<=1})
   %
   % We compute dt for flux_in = darcy flux and flux_in = flux due to
   % gravity, and use min(dt1, dt2).

   if opt.ComputeDt,
      dt = opt.max_dt;
      s  = linspace(0, 1, 1001) .';

      deriv = @(f) diff(f(fluid.mu, fluid.kr(struct('s', s)))) ./ diff(s);

      if nnz(gravmat),
         f  = @(mu,kr) (prod(kr,2) ./ prod(mu,2)) ./ ...
                        sum(bsxfun(@rdivide, kr, mu), 2);

         dt = min(dt, estimate_dt(gravmat, deriv(f), pv, 0));
      end
      if ~opt.OnlyGrav,
         f  = @(mu,kr) (kr(:,1) ./ mu(1)) ./ ...
                        sum(bsxfun(@rdivide, kr, mu), 2);

         dt = min(dt, estimate_dt(flux, deriv(f), pv, q));
      end

      dt = dt * opt.dt_factor;
   else
      % Return dt = 0 if no estimate of timestep is requested.
      dt = 0;
   end
end

%--------------------------------------------------------------------------
% Helpers follow.
%--------------------------------------------------------------------------

function [gravity_matrix, flux] = matrices_nograv(G, resSol, is_int)
% matrices_nograv -- Compute flux and (trivial) gravity matrix in absence
%                    of gravity effects.
   nc = G.cells.num;

   % - Gravity matrix -
   gravity_matrix = sparse(G.cells.num, sum(double(is_int)));

   % - Flux matrix -
   % resSol.cellFlux is amount of outflow for all half contacts on all
   % cells.  Negative cell flux is flux into a cell.
   %
   is_inflow    = resSol.cellFlux < 0;

   inflow_faces = G.cellFaces(is_inflow, 1);
   inflow_flux  = resSol.cellFlux(is_inflow);

   N            = double(G.faces.neighbors(inflow_faces, :));
   cellNo       = rldecode(1:G.cells.num, double(G.cells.numFaces), 2)';
   sgn      = 2*double(G.faces.neighbors(G.cellFaces(:,1), 1)==cellNo)-1;

   ind          = sgn(is_inflow) > 0;

   N(ind,[1,2]) = N(ind,[2,1]);

   % In distributing the flow pattern, only consider internal faces when
   % constructing the connection matrix.  Flux boundary conditions, if any,
   % are considered sources and handled separately (function 'contrib_bc').
   %
   int_face = prod(N, 2) > 0;
   int_conn = N(int_face, :);
   int_flux = inflow_flux(int_face);

   flux = sparse(int_conn(:,2), int_conn(:,1), int_flux, nc, nc);
end

%--------------------------------------------------------------------------

function [gravity_matrix, flux] = matrices(G, rock, fluid, resSol, ...
                                           grav_only, is_int)
% matrices -- Compute flux and gravity matrix in presence of gravity.
% is_int - Logical vector of length G.faces.num, true for internal face.

   [nc, nf] = deal(G.cells.num, sum(double(is_int)));
   cellNo   = rldecode(1 : nc, double(G.cells.numFaces), 2) .';

   % Indices to (internal) half-faces.
   cIntFInx = is_int(G.cellFaces(:,1));

   % Global-to-local face map for internal faces.
   G2L      = cumsum(double(is_int));

   % Indices of internal face corresponding to cIntFInx.
   globfix  = G.cellFaces(cIntFInx, 1);

   % Renumbering globfix to 1:numel(globfix)
   locfix   = G2L(globfix);

   % - Gravity matrix -
   % Spread gravity to cellfaces: cellFInx * g_const .* cellFace_normal

   % rho(1) - rho(2) is always correct (even when rho(2) > rho(1)), because
   % during transport (using, e.g., the 'twophaseUpwBEGrav' transport
   % solver) we only modify the *first* saturation component.
   g   = gravity() * (fluid.rho(1) - fluid.rho(2));
   dim = size(G.nodes.coords, 2);

   harm          = zeros([G.faces.num, 1]);
   renum         = zeros([G.faces.num, 1]);
   renum(is_int) = 1 : nf;

   % nKg == n' * K * g for all cellfaces.
   [K, r, c] = get_perm(rock, dim);
   nKg = sum(G.faces.normals(G.cellFaces(:,1), r) .* ...
             bsxfun(@times, K(cellNo,:), g(c)), 2);

   % Compute harmonic average of nKg on all *internal* faces.
   harm(is_int) = 1 ./ accumarray(renum(globfix), 1 ./ nKg(cIntFInx));

   sgn = 2*double(G.faces.neighbors(G.cellFaces(:,1), 1) == cellNo) - 1;

   % Compute final gravity matrix.
   gravity_matrix = sparse(cellNo(cIntFInx), locfix, ...
                           sgn(cIntFInx) .* harm(globfix), nc, nf);

   % - Flux matrix -
   if grav_only,
      flux = sparse(G.cells.num, nf);
   else
      flux = sparse(cellNo(cIntFInx), locfix, ...
                    resSol.cellFlux(cIntFInx), nc, nf);
   end
end

%--------------------------------------------------------------------------

function dt = estimate_dt(flux_in, df, pv, q)
   % Consider max of inflow or outflow from a cell
   % when estimating time step size.

   if size(flux_in,2) == numel(q)
      flux_out = abs(sum(flux_in,1))' - min(q,0);
      flux_in  = abs(sum(flux_in,2))  + max(q,0);
   else
      flux_out = flux_in;
      flux_out(flux_out < 0) = 0;
      flux_out = abs(sum(flux_out, 2)) - min(q,0);

      flux_in(flux_in > 0) = 0;
      flux_in = abs(sum(flux_in, 2)) + max(q,0);
   end
   flux_max = max(flux_in, flux_out);
   is_flux = flux_max ~= 0;

   dt = min((pv(is_flux) ./ flux_max(is_flux)) ./ max(abs(df)));
end


%--------------------------------------------------------------------------
%
% Source term contributions from wells, boundary conditions and explict
% sources.
%
%--------------------------------------------------------------------------
%
% We need special treatment for injection into reservoir.  Specifically, we
% need to take into account the fluid composition of the injected fluid
% lest the subsequent transport solve (using e.g. 'twophaseUpwFE') be
% thoroughly confused.
%
% For wells, this composition is available in the field '.comp_i' (one
% saturation, [Aqua, Liquid, Vapor], for each *well*), while explicit
% sources provide one three-component saturation value for each
% injection *cell* (field 'src.sat').

function [qi, qs] = contrib_wells(W, wellSol)
   % Wells as defined by 'addWell'.

   nperf = cellfun(@numel, { W.cells }) .';

   qi = vertcat(W.cells);
   qs = vertcat(wellSol.flux);

   % Injection perforations have positive flux (into reservoir).
   %
   comp      = rldecode(vertcat(W.compi), nperf);
   inj_p     = qs > 0;
   qs(inj_p) = qs(inj_p) .* comp(inj_p,1);
end

%--------------------------------------------------------------------------

function [qi, qs] = contrib_src(src)
   % Explicit sources defined by (e.g.) 'addSource'.

   qi = src.cell;
   qs = src.rate;

   % Injection sources have positive rate into reservoir.
   %
   in = find(src.rate > 0);
   if ~isempty(in),
      qs(in) = qs(in) .* src.sat(in,1);
   end
end

%--------------------------------------------------------------------------

function [qi, qs] = contrib_bc(G, resSol, bc, is_int)
   % Contributions from boundary conditions as defined by 'addBC'.

   qs = zeros([G.faces.num, 1]);
   dF = false([G.faces.num, 1]);

   isDir = strcmp('pressure', bc.type);
   isNeu = strcmp('flux',     bc.type);

   dF(bc.face(isDir))      = true;
   cfIx                    = dF(G.cellFaces(:,1));

   qs(G.cellFaces(cfIx,1)) = -resSol.cellFlux(cfIx);
   qs(bc.face(isNeu))      = bc.value(isNeu);

   % Injection BC's have positive rate (flux) into reservoir.
   %
   is_inj = qs > 0;
   if any(is_inj),
      qs(is_inj) = qs(is_inj) .* bc.sat(is_inj(bc.face), 1);
   end

   is_outer = ~is_int;

   qi = sum(G.faces.neighbors(is_outer,:), 2);
   qs = qs(is_outer);
end

%--------------------------------------------------------------------------

function [K, r, c] = get_perm(rock, dim)
   [nc, nk] = size(rock.perm);

   if dim == 3,
      switch nk,
         case 1,
            K = rock.perm * [1, 0, 0, 1, 0, 1];
         case 3,
            K = [rock.perm(:,1), zeros([nc,2]), ...
                 rock.perm(:,2), zeros([nc,1]), ...
                 rock.perm(:,3)];
         case 6,
            K = rock.perm;
         otherwise,
            error('initTransport:PermDim:Unsupported', ...
                 ['%d-component permeability value is not supported ', ...
                  'in three space dimensions.'], nk);
      end

      K = K(:, [1, 2, 3, 2, 4, 5, 3, 5, 6]);
      r =      [1, 1, 1, 2, 2, 2, 3, 3, 3] ;
      c =      [1, 2, 3, 1, 2, 3, 1, 2, 3] ;

   else
      error('initTransport:PhysDim:Unsupported', ...
            '%d space dimensions are unsupported at this time.', dim);
   end
end
