function resSol = twophaseUpwBE(resSol, tf, q, gm, pv, fluid, varargin)
%Implicit single point upwind solver for two-phase flow, no gravity.
%
% SYNOPSIS:
%   resSol = twophaseUpwBE(resSol, t, q, gm, porvol, fluid)
%   resSol = twophaseUpwBE(resSol, t, q, gm, porvol, fluid, 'pn1', pv1, ...)
%
% DESCRIPTION:
%   Function twophaseUpwBE solves the Buckley-Leverett transport equation
%
%        s_t + f(s)_x = q
%
%   using a first-order upwind discretisation in space and a backward Euler
%   discretisation in time.  The nonlinear system of equations that must be
%   solved to move the solution from time=0 to time=tf, are solved using a
%   Newton-Raphson algorithm with line search to increase robustness.
%
%   In the case of failing to compute a solution using only a single step
%   over the interval [0,tf], an alternative strategy involving sub-steps
%   and step size control is employed.
%
% REQUIRED PARAMETERS:
%   resSol   - Reservoir solution structure containing valid (water)
%              saturation resSol.s(:,1) with one value for each cell in the
%              reservoir model.
%
%   tf        - Length of time step, measured in units of seconds.
%
%   q        - Accumulated sources (typically contributions from wells).
%              One scalar value for each cell in the grid.  The source
%              rates are assumed to be measured in units of m^3/s.
%
%   gm       - Upwind/inflow matrix of fluxes into each cell, created
%              (e.g.) by function initTransport.  Specifically, gm(i,j) is
%              the (positive) flux from cell j to cell i.  The fluxes are
%              assumed to be measured in units of m^3/s.
%
%   porvol   - Reservoir pore volumes, measured in units of m^3, one scalar
%              value for each cell in the reservoir model.
%
%   fluid    - Data structure describing the fluids in the problem.
%
% OPTIONAL PARAMETERS (supplied in 'key'/value pairs ('pn'/pv ...)):
%
%   verbose  - Whether or not time integration progress should be reported
%              to the screen.
%              Default value: verbose = false.
%
%   nltol    - Absolute tolerance of iteration.  The numerical solution
%              must satisfy the condition
%
%                NORM(S-S0 + dt/porvol(out - in) - Q, INF) <= nltol
%
%              at all times in the interval [0,t].
%
%              Default value: nltol = 1.0e-6.
%
%   lstrials - Maximum number of trials in linesearch method.  Each new
%              trial corresponds to halving the step size along the search
%              direction.
%              Default value: lstrials = 20.
%
%   maxnewt  - Maximum number of inner iterations in Newton-Raphson method.
%              Default value: maxnewt = 25.
%
%   tsref    - Maximum time step refinement power.  The minimum time step
%              allowed is t / 2^tsref.
%              Default value: tsref = 12.
%
% RETURNS:
%   resSol   - Reservoir solution with updated resSol.s.
%
% SEE ALSO:
%   twophaseUpwBEGrav, initTransport, implicitTransport, explicitTransport.

%{
Copyright 2009 SINTEF ICT, Applied Mathematics.

This file is part of The MATLAB Reservoir Simulation Toolbox (MRST).

MRST is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MRST is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MRST.  If not, see <http://www.gnu.org/licenses/>.
%}

% $Id: twophaseUpwBE.m 2138 2009-05-04 17:09:29Z bska $

[F, Jac, linsrch, verbose, nltol, maxnwt, mints] = ...
   setup(tf, q, gm, pv, fluid, varargin{:});

sz = size(resSol.s(:,1));
sn = resSol;

printif(verbose, '\n\n');
printif(verbose, [repmat('-', [1, 70]), '\n']);
printif(verbose, '  Time interval           iter   relax    residual \n');
printif(verbose, [repmat('-', [1, 70]), '\n']);

[t, dt, count] = deal(0.0, tf, 0);

%--------------------------------------------------------------------------
% Main time loop (ideally, just a single step) ----------------------------
%
if verbose, tic, end
while t < tf && dt > mints,
   dt = min(dt, tf - t);

   %-----------------------------------------------------------------------
   % Outer controlling loop (sub step size &c) ----------------------------
   %
   redo_newton = true;
   while redo_newton,
      printif(verbose, '[%f, %f]:', t, t + dt); nc = 0;

      s0 = resSol.s(:,1);
      sn = resSol;

      %--------------------------------------------------------------------
      % Initialise NR algorithm -------------------------------------------
      %
      res     = F(sn, s0, dt);
      err     = norm(res, inf);
      nwtfail = err > nltol;
      linfail = false;
      it      = 0;
      %--------------------------------------------------------------------
      % Execute inner NR algorithm ----------------------------------------
      %
      while nwtfail && ~linfail && it < maxnwt,
         ds = - Jac(sn, dt) \ res;
         [sn, res, alph, linfail] = linsrch(sn, s0, ds, dt, err);

         it      = it + 1;
         err     = norm(res, inf);
         nwtfail = err > nltol;

         printif(verbose, repmat('\b', [1, nc]));
         nc = printif(verbose, '\t%4d\t %2.2f\t %5.5e', it, alph, err);
      end

      %--------------------------------------------------------------------
      % Determine if NR succeeded or not ----------------------------------
      %
      count = count - 1;
      if nwtfail,
         printif(verbose, '\tReducing step.');

         dt    = dt / 2;
         count = 5;
      else
         redo_newton = false;
         t = t + dt;
         if it == 0,
            printif(verbose, '\t%4d\t %s\t %5.5e', it, '-', err);
            printif(verbose, '\t NB: err <= ntol.');
         end

         if it <= 5  && count < 0 && t ~= tf, % Arbitrary threshold
            printif(verbose, '\tIncreasing step.');

            dt = min(1.5 * dt, tf - t);
         end
      end
      printif(verbose, '\n');
   end

   % Time step [t, t+dt] completed successfully.
   resSol = sn;
end
tocif(verbose)

if ~(t < tf),
   printif(verbose, 'We''re done (%f)\n', t);
else
   printif(verbose, 'Unable to integrate to final time, t=%f\n', tf);
   resSol.s(:,1) = NaN(sz);
end

if size(resSol.s,2) > 1,
   % Update oil saturation:
   resSol.s(:,2) = 1 - resSol.s(:,1);
end
end

%--------------------------------------------------------------------------
% Private helpers follow.
%--------------------------------------------------------------------------

function [F, Jac, linsrch, verbose, ...
      nltol, maxnwt, mints] = setup(tf, q, gm, pv, fluid, varargin)

nc = numel(pv);

if ~((ndims(gm) == 2) && all(size(gm) == nc)),
    error('twophaseUpwBE:SETUP:WRONG_SIZE',           ...
         ['Unexpected size of inflow flux matrix.\n', ...
          'Expected [', int2str([nc, nc]), '],\n',    ...
          'but got  [', int2str(size(gm)), ']']);
end

prm  = struct(  'verbose',  false,  ...  % emit progress reports
                'nltol',    1.0e-6, ...  % non-linear residual tolerance
                'lstrials', 20,     ...  % max no of line search trials
                'maxnewt',  25,     ...  % max no. of NR iterations
                'tsref',    12,     ...  % time step refinement
                'resred',   0.99);       % residual reduction factor

prm = merge_options(prm, varargin{:});

verbose = prm.verbose;
nltol   = prm.nltol;
maxnwt  = prm.maxnewt;
mints   = pow2(tf, -prm.tsref);        % minimum time step size

[im, jm, valm] = find(gm);

q  = q (:);
pv = pv(:);

% System F(s) = 0 of non-linear equations (and the system's Jacobian
% matrix) defining saturation equilibrium at a given time step.
%
% Here, we use a Buckley-Leverett model with
%
%            s²/µw              s²                µw
%    f(s) = ---------------- = ------------ ,  mr=---
%            s²/µw+(1-s)²/µo    s²+mr*(1-s)²      µo
%
%
%            2 mr*s(1-s)
%   df(s) = ---------------
%           (s²+mr*(1-s)²)²
%
%
% With the matrix v, we can write an upwind backward Euler
% discretisation of the Buckley-Leverett model as
%
%    s^n-s^{n-1} + dt*v*f(s) = MAX(q,0),
%
% where
%
%    v(i,j) = - flux into cell i from cell j,
%    v(i,i) = sum of flux out of cell i -MIN(q(i),0).
%
% The target function is
%
%    F(s) = s-s^{n-1} + dt*v*fs(s) - max(q, 0)
%
% and the Jacobian is
%
%   dF(s) =  I + dt*v*diag(df(s)).
%
f    = @(kr) (  kr(:,1)./(kr(:,1) + (fluid.mu(1)./fluid.mu(2))*kr(:,2)) );
df_w = @(kr, dkr, Lt) ((dkr(:,1)./fluid.mu(1))./Lt - ...
                (kr(:,1)./fluid.mu(1)).* ...
                sum(bsxfun(@rdivide, dkr, fluid.mu),2)./(Lt.^2));

% v <- (gm - colsum(gm) + MIN(q,0))) / pore_volume
v = spdiags(1./pv, 0, nc, nc) * sparse( [  im;    jm;  (1 : nc)'], ...
                                        [  jm;    jm;  (1 : nc)'], ...
                                        [valm; -valm; -min(q,0) ], nc, nc);

F   = @(sol, s0, dt) (sol.s(:,1) - s0 + dt.* ...
                     (v*f(fluid.kr(sol)) - max(q./pv,0)));
Jac = @(sol, dt) ( speye(nc) + dt.*v*sparse(1:nc, 1:nc, ...
                  df_w( fluid.kr(sol), fluid.dkr(sol), fluid.Lt(sol) ), ...
                  nc, nc, nc) );


% Bind prm.resred * err, @(sat) F(sat, s0, dt) and prm.lstrials
% to function call to reduce clutter in main code.
linsrch = @(sol, s0, ds, dt, err) linesearch(sol, ds, prm.resred * err, ...
                                             @(sol) F(sol, s0, dt),     ...
                                             prm.lstrials);
end

%--------------------------------------------------------------------------

function [sol, res, alph, fail] = linesearch(sol, ds, target, F, ni)
%
% Basic idea: search for a step size 'alpha' in direction 'ds', subject to
% the restriction that alpha be in [0,1], which ensures that the objective
% function 'F' decreases.  That is: F(s + alpha*ds) < F(s).
%
% In the current implementation, alpha is reduced in a geometric sequence.
% A more sophisticated approach would ensure a certain minimum reduction as
% well.
minSat = 0;     % Minimum (water) saturation
maxSat = 1;     % Maximum (water) saturation
capSat = @(sat) min(max(minSat, sat), maxSat);

alph = 0;
i    = 0;
fail = true;

% Geometric line search: seems pretty robust
while fail && (i < ni),
   sn.s   = capSat(sol.s(:,1) + pow2(ds, alph));
   res  = F(sn);

   alph = alph - 1;
   i    = i + 1;
   fail = ~(norm(res, inf) < target);
end

alph       = pow2(alph + 1);      % Undo last (unneeded) scaling.
sol.s(:,1) = sn.s;
end

%--------------------------------------------------------------------------

function nc = printif(flag, varargin)
if flag,
    nc = fprintf(varargin{:});
else
    nc = 0;
end
end
