function grdecl = readGRDECL(fn, varargin)
%Read Eclipse file assuming corner-point grid format.
%
% SYNOPSIS:
%   grdecl = readGRDECL(fn)
%   grdecl = readGRDECL(fn, 'pn1', pv1, ...)
%
% PARAMETERS:
%   fn - String holding name of (readable) GRDECL specification.
%        Note: The GRDECL specification is assumed to be a physical
%        (seekable) file on disk, not to be read in through (e.g.)
%        a POSIX pipe.
%
%   'pn'/pv - List of 'key'/value pairs designating optional parameters.
%             Currently supported values are
%               - verbose  -- Emit messages to screen while processing.
%                             Default value: FALSE.
%
%               - usecache -- Cache result to mat-file to avoid reading
%                             grid file when it has not changed.
%                             Default value: TRUE.
%
% RETURNS:
%   grdecl - Output structure containing the known, though mostly
%            unprocessed, fields of the GRDECL specification.
%
%            With the exception of keywords regarding wells, stored in
%            'grdecl' structure field 'wells', and 'SPECGRID' whose first
%            three arguments (grid cell dimensions NX, NY, NZ) are stored
%            in 'grdecl' structure field 'cartDims', all currently
%            recognized GRDECL keywords are stored in equally named
%            'grdecl' structure fields. Specifically, GRDECL keyword
%            'COORD' is stored 'grdecl' structure field 'COORD' and so
%            forth.
%
%            The pillar description 'COORD' is stored in an nPillar-by-6
%            array (number of pillars, nPillar == (NX+1)*(NY+1)) of
%            bottom/top coordinate pairs.  Specifically,
%
%              grdecl.COORD(i,1:3) -- (x,y,z)-coordinates of pillar 'i'
%                                     top point.
%              grdecl.COORD(i,4:6) -- (x,y,z)-coordinates of pillar 'i'
%                                     bottom point.
%
%            The currently recognized GRDECL keywords are 'SPECGRID',
%            'COORD', 'ZCORN', 'PORO', 'PERMX', 'PERMY', 'PERMZ', 'ACTNUM',
%            'WELSPECS', 'COMPDAT', 'WCONINJE','WCONPROD'.
%
% SEE ALSO:
%   processGRDECL, readWellKW. 
%

%{
Copyright 2009 SINTEF ICT, Applied Mathematics.

This file is part of The MATLAB Reservoir Simulation Toolbox (MRST).

MRST is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MRST is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MRST.  If not, see <http://www.gnu.org/licenses/>.
%}

% $Id: readGRDECL.m 1956 2009-03-31 11:06:38Z bska $

opt = struct('verbose', false, ...
             'usecache', true);
opt = merge_options(opt, varargin{:});
verbose  = opt.verbose;
%usecache = opt.usecache;


[fid, msg] = fopen(fn, 'rt');
if fid < 0, error(msg), end

cartDims        = [-1, -1, -1];
numCell         = -1;
grdecl.TSTEP    = [];
grdecl.PSIDE    = struct([]);
grdecl.FLUXSIDE = struct([]);
grdecl.NOGRAV   = false;
grdecl.METRIC   = true;
grdecl.WATER    = true;
grdecl.OIL      = true;
grdecl.GAS      = false;
%grdecl.DISGAS   = false;

grdecl.wells = initDefaultWell();
                 
while ~feof(fid),
   lin = fgetl(fid);
   if lin == -1,
      msg = ferror(fid, 'clear');
      fclose(fid);
      error('readGRDECL:Input:Empty', ...
            'GRDECL file `%s'' is unreadable.\nSystem reports: %s\n', ...
            fn, msg);
   end

   % Loop until next keyword
   kw = regexp(lin, '^[A-Z]+', 'match', 'once');
   while isempty(kw) && ~feof(fid),
      lin = fgetl(fid);
      if lin ~= -1,
         kw = regexp(lin, '^[A-Z]+', 'match', 'once');
      end
   end

   if ~feof(fid),
      dispif(verbose, 'Reading keyword ''%s''\n', kw);
      switch kw,
         case {'SPECGRID','DIMENS'}
            cartDims = fscanf(fid, '%f', 3) .';
            numCell  = prod(cartDims);
            trash    = fgetl(fid); %#ok
            grdecl.cartDims = cartDims(:) .';

         case 'COORD',
            checkDim(cartDims, numCell, kw);
            grdecl.COORD = readVector(fid, kw, ...
                                      6 * prod(cartDims(1:2) + 1));

         case 'ZCORN',
            checkDim(cartDims, numCell, kw);
            grdecl.ZCORN = readVector(fid, kw, 8 * numCell);

         case {'PORO',                       ...
               'PERMX',  'PERMXY', 'PERMXZ', ...
               'PERMYX', 'PERMY',  'PERMYZ', ...
               'PERMZX', 'PERMZY', 'PERMZ',  ...
               'ACTNUM'},
            checkDim(cartDims, numCell, kw);
            grdecl.(kw) = readVector(fid, kw, numCell);

         case {'COORDX','COORDY','COORDZ'},
            checkDim(cartDims,numCell,kw);
            grdecl.(kw) = readVector(fid, kw, prod(cartDims+1));

         case {'NOGRAV', 'METRIC', 'WATER', 'OIL', 'GAS'}
            grdecl.(kw) = true;

         case 'FIELD'
            grdecl.METRIC = false;

         case {'PSIDE', 'FLUXSIDE'}
            lin    = fgetl(fid);
            tokens = regexp(lin, '\S+', 'match');

            if strcmp(tokens{1}, 'GLOBAL') && ...
               any(strcmp(tokens{2}, ...
                          {'LEFT', 'RIGHT', ...
                           'FRONT', 'BACK', ...
                           'TOP', 'BOTTOM'})),
               vals = str2num(strvcat(tokens{3:end-1})); %#ok

               if numel(vals) == 5,
                  grdecl.(kw) = [grdecl.(kw), struct('direction', tokens{2}, ...
                                                     'value', vals(1),       ...
                                                     'range', vals(2:end))];
               else
                  error([kw, ':SPEC:Incorrect'], ['Incorrect ', kw, ' Spec found']);
               end
            else
               error([kw, ':SPEC:Incorrect'], ['Incorrect ', kw, ' Spec found']);
            end

         case 'DXV',
            % Error checking?
            checkDim(cartDims, numCell, kw);
            grdecl.(kw) = readVector(fid, kw, cartDims(1));

         case 'DYV',
            checkDim(cartDims, numCell, kw);
            grdecl.(kw) = readVector(fid, kw, cartDims(2));

         case 'DZV',
            checkDim(cartDims, numCell, kw);
            grdecl.(kw) = readVector(fid, kw, cartDims(3));

         case 'DEPTHZ',
            checkDim(cartDims, numCell, kw);
            grdecl.(kw) = readVector(fid, kw, prod(cartDims(1:2) + 1));

         case 'TSTEP',
            grdecl.TSTEP = [grdecl.TSTEP; readVector(fid, kw, inf)];

         case {'SWAT', 'SGAS', 'PRESSURE', 'RS', 'RV'}
            checkDim(cartDims, numCell, kw);
            grdecl.(kw)  = readVector(fid, kw, numCell);                                      
         
         case { 'WELSPECS', 'COMPDAT', 'WCONINJE','WCONPROD'}
            grdecl.wells = readWellKW(fid, grdecl.wells, kw);    
                  
         otherwise,
            warning('readGRDECL:KEYWORD:Unrecognized', ...
                    'Input keyword ''%s'' is unrecognized\n', kw);
      end
   end
end
fclose(fid);

%remove default well
grdecl.wells(1) = [];

%--------------------------------------------------------------------------
% Private helpers follow.
%--------------------------------------------------------------------------

function well = initDefaultWell() 
% Initialize the default well structure
% NOTE:
% * all fields ending with '_vec' are vectors with values for each COMPDAT.
% * for comments define: rate_trg = rate target, up_lim = upper limit.
   
   well = struct( ...
         'name',       '', ...
         'I',          0,  ...
         'J',          0,  ...
         'I_vec',      0, ...
         'J_vec',      0, ...
         'K_vec_u',    -1, ... %Location of upper connecting block.
         'K_vec_l',    -1, ... %Location of lower connecting block.
         'state_vec',   1, ... %OPEN/SHUT flag of connection, 1 = 'OPEN'.
         'diam_vec',   0.25, ...
         'transm_vec', -1, ... %Transm. factor of connection (WI).
         'Kh_vec',     -1, ... %Permeability thickness.
         'skin_vec',   0.0,... %Skin factor of connection.
         'dir_vec',    'Z',... %Dir of wellblock penetration: X, Y or Z.
         'state',      'OPEN', ... %OPEN or SHUT well.
         'control',    '', ... %rat_trg type or BHP.
         'oil_rate',   [], ... %rat_trg/up_lim (prod).
         'water_rate', [], ... %rat_trg/up_lim (prod).
         'gas_rate',   [], ... %rat_trg/up_lim (prod).
         'liquid_rate',[], ... %rat_trg/up_lim (prod) or surface rate (inj).
         'vol_rate',   [], ... %rat_trg/up_lim (reservoir fluid vol-rate).
         'bhp',        [], ... %bhp target or lower limit.
         'type',       '', ... %Injection/Production.
         'inj_type',   '', ... %WAT, GAS, OIL.
         'compdat_count', 0);  %Count number of compdats.

function checkDim(cartDims, numCell, kw)
if numCell < 1 || any(cartDims < 1),
   error('readGRDECL:Input:NoDim', ...
         'GRDECL keyword ''%s'' found before dimension specification', kw);
end
