function v = readVector(fid, field, nel)
%Input vector of floating point numbers from GRDECL file.
%
% SYNOPSIS:
%   v = readVector(fid, field, nel)
%
% PARAMETERS:
%   fid   - File identifier (as defined by FOPEN) of GRDECL file open for
%           reading.  Assumed to point to a seekable (i.e. physical) file
%           on disk, not (e.g.) a POSIX pipe.
%
%   field - Name (string) identifying the GRDECL keyword (field) currently
%           being processed.  Used for error identification/messages only.
%
%   nel   - Number of elements to read from input stream.  As a special
%           case, the caller may pass nel==INF (or nel=='inf') to read as
%           much as possible.  In this case it is imperative that the input
%           vector be terminated by a '/' character.
%
% RETURNS:
%   v     - Vector, length 'nel', of floating point numbers defining the
%           contents of GRDECL keyword 'field'.  If nel==INF, NUMEL(v) is
%           the number of vector elements read before the terminating slash
%           character.

%{
Copyright 2009 SINTEF ICT, Applied Mathematics.

This file is part of The MATLAB Reservoir Simulation Toolbox (MRST).

MRST is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MRST is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MRST.  If not, see <http://www.gnu.org/licenses/>.
%}

% $Id: readVector.m 1956 2009-03-31 11:06:38Z bska $

% Simple vector read.  Assume the vector is given explictly by listing all
% of its values separated by whitespace, possibly interspersed by comments.
%

if ischar(nel) && strcmp(nel, 'inf'), nel = inf; end
opt = getScanOpt(nel);
C   = textscan(fid, '%f', opt{:});

pos = ftell(fid);
lin = fgetl(fid);
while ischar(lin) && (isempty(lin) || matches(lin, '^\*|^\s+$')),
   if strmatch('*', lin),
      % Repeat description of the form
      %      N*value
      % with 'N' a positive integer.  There must be no whitespace
      % immediately before or after the literal asterisk ('*').
      %
      % This format allows considerable economy of file space, e.g. when
      % listing uniform porosity values such as
      %
      %   PORO
      %     5000*0.3 /
      %
      % for a vector of five thousand repetitions of the value 0.3.  The
      % MATLAB equivalent is
      %
      %   PORO = REPMAT(0.3, [5000, 1]);
      %
      % This format is most likely encountered when processing GRDECL
      % keywords such as 'DXV', 'PERMX', 'PORO', 'ACTNUM' and, possibly,
      % 'ZCORN' and 'TSTEP'.
      %

      % Validate repeat description format
      fseek(fid, pos - 1, 'bof');
      repdesc = fscanf(fid, '%c', 3);
      if numel(repdesc) ~= 3 || ~matches(repdesc, '\d\*\d'),
         error('readGRDECL:RepeatDescr:Malformed',                    ...
               'Incorrect repeat description detected in keyword %s', ...
               field);
      end

      % Validate repeat count value.
      nrep = C{1}(end);
      if ~isnumeric(nrep)             || ...
         nrep < 1                     || ...
         nrep > nel - numel(C{1}) + 1 || ...
         nrep ~= fix(nrep),
         error('readGRDECL:RepeatCount:OutOfBounds',             ...
               ['Unexpected repeat count in keyword %s.\n',      ...
                'Expected integral repeat count in [1 .. %d], ', ...
                'but found %f.'], field, nel - numel(C{1}) + 1, nrep);
      end

      % Append repeated value to vector.
      [val, count, errmsg, nextix] = sscanf(lin(2:end), '%f', 1);
      C   = {[C{1}(1:end-1); repmat(val, [nrep, 1])]};
      nv  = numel(C{1});

      % Reposition reader at first whitespace or '/' character whereupon
      % regular vector reading may resume.
      fseek(fid, pos + nextix, 'bof');

      opt = getScanOpt(nel, nv);
      C1 = textscan(fid, '%f', opt{:});
      C  = {[C{1}(:); C1{1}(:)]};
   end
   pos = ftell(fid);
   lin = fgetl(fid);
end

iseof           = feof(fid);
[errmsg, errno] = ferror(fid);
readAll = (~isinf(nel) && (numel(C{1}) == nel)) | isinf(nel);
if ischar(lin) && matches(lin, '^\s*/') && readAll,
   v = C{1};
elseif ischar(lin) && matches(lin, '^\s*/'),
   fclose(fid);
   error(['readGRDECL:Vector_', field, ':PrematureTermination'],    ...
         ['Detected termination character ''/'' at position %d.\n', ...
          'Scanning of keyword ''%s'' not completed.'], pos, field);
elseif ischar(lin) && matches(lin, '^[A-Z]+'),
   fclose(fid);
   error(['readGRDECL:Vector_', field, ':OtherKeyword'],           ...
         ['Encountered non-numeric data ''%s'' at position %d.\n', ...
          'Scanning of keyword ''%s'' not completed.\n',           ...
          'Missing slash (/) in input?'],                          ...
          lin, pos, field);
elseif ischar(lin) && matches(lin, '^[a-z!;=]+'),
   error(['readGRDECL:Vector_', field, ':Unexpected'],             ...
         ['Encountered non-numeric data ''%s'' at position %d.\n', ...
          'Scanning of keyword ''%s'' not completed.'],            ...
          lin, pos, field);
elseif iseof,
   fclose(fid);
   error(['readGRDECL:Vector_', field, ':EOF'],               ...
         'End of file before complete read of keyword ''%s''.', field);
elseif errno ~= 0,
   fclose(fid);
   error(['readGRDECL:Vector_', field, ':SystemError'], ...
         ['Read error while processing keyword %s.\n',  ...
          'Input system reports ''%s''.'], field, errmsg);
else
   fclose(fid);
   error(['readGRDECL:Vector_', field, ':Malformed'],     ...
         ['Unexpected input while reading keyword %s.\n', ...
          'Found ''%s'' at position %d.'], field, lin, pos);
end

%--------------------------------------------------------------------------
% Private helpers follow.
%--------------------------------------------------------------------------

function opt = getScanOpt(nel, varargin)
opt = {'CommentStyle', '--'};
if nargin > 1 && isnumeric(varargin{1}),
   nel = nel - varargin{1};
end
if ~isinf(nel),
   opt = {nel, opt{:}};
end

%--------------------------------------------------------------------------

function b = matches(str, pat)
b = ~isempty(regexp(str, pat, 'once'));
